/*
 * bushfire_algorithm.h
 *
 * Template based implementation of a brushfire algorithm (bushfire since we are
 * in Aus) for a 2-dimensional grid.
 * - Each bushfire2 instance represents a type of wavefront.
 * - The priority queue provides the facility to get the next element within
 *   the wavefront.
 * - This is simpler than the bushfire class which tries to provide multiple
 *   wavefront types but ends up being too complicated.
 *
 *  Created on: 13/08/2011
 *      Author: daver
 */

#ifndef BUSHFIRE2_H_
#define BUSHFIRE2_H_

#include <queue>
#include <utility>
#include <boost/assert.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/foreach.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>

// Metric is the metric for the queue (usually int or float)
// Descriptor is the type of object we are interested in (eg. graph vertex descriptor or a point)
template <typename Metric, typename Descriptor>
class PriorityQueueComparison
{
  bool reverse_;
public:
  PriorityQueueComparison(bool rev=false): reverse_(rev){}
  typedef std::pair<Metric, Descriptor> t_pair;
  bool operator()(const t_pair& lhs, const t_pair& rhs) const
  {
    if (reverse_) return lhs.first > rhs.first;
    return lhs.first < rhs.first;
  }
};

template<typename Metric,
         typename Descriptor,
         typename Compare=PriorityQueueComparison<Metric,Descriptor> >
class BushFire2
{
public:
  typedef BushFire2<Metric,Descriptor,Compare> t_bushfire2;
  typedef boost::function<void (t_bushfire2&, const Metric&, const Descriptor&)> t_action;

private:
  typedef std::pair<Metric, Descriptor> t_element;
  typedef std::priority_queue<std::pair<Metric, Descriptor>, std::vector<t_element>, Compare> t_queue;
  bool reverse_;
  t_queue queue_;
  bool done_;
public:
  BushFire2(){}
  BushFire2(bool reverse): reverse_(reverse), queue_(Compare(reverse_)){}

  /* Add an element to the bushfire */
  void push(const Metric& prio, const Descriptor& desc)
  {
    queue_.push(make_pair(prio, desc));
  }

  void done(){ done_ = true; }

  void clear() { queue_ = t_queue(Compare(reverse_)); }

  size_t size() const { return queue_.size(); }

  void update(t_action action)
  {
    done_ = false;
    while (!queue_.empty() && !done_)
    {
      t_element top = queue_.top();
      queue_.pop();
      action(*this, top.first,top.second);
    }
  }

  template<typename C, typename Fn>
  static t_action make_callback(C fo, Fn fn)
  {
    return boost::bind(fn,fo,_1,_2,_3);
  }
};




#endif /* BUSHFIRE2_H_ */
