/*
 * connected_points.h
 *
 *  Created on: 30/09/2011
 *      Author: daver
 */

#ifndef CONNECTED_POINTS_H_
#define CONNECTED_POINTS_H_

#include <map>
#include <tr1/unordered_map>
#include <boost/foreach.hpp>
#include <boost/functional/hash.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/astar_search.hpp>

//#include <boost/graph/connected_components.hpp>


#include <boost/shared_ptr.hpp>
#include "mapmetadata_tools.h"

// A distance metric class for calculating the distance along a path.
// Tracks movements sideways (horizontal and vertical) as well
// as diagonal. Comparing metrics can be done precisely by resorting
// to floating point numbers, or approximately using an integers
// only. The total distance travelled is the distance travelled in
// sideways movements and the distance travelled in diagonal movements.
// The distance travelled by a single sideway movement is simply 1 unit,
// while a diagonal movement is the square root of 2. Therefore the
// total distance is s + d*sqrt(2).
// The square of the distance travelled on a path is:
//    (s+ d*sqrt(2))^2 =  s^2 + 2*s*d*sqrt(2) + 2*d^2. I approximate sqrt(2)
// as 1.5 and therefore 2*1.5 is 3. So the approximate square of the distance
// is s^2 + 3*s*d + 2*d^2.

/*
class DistMetric
{
public:
  int sideways_;
  int diagonals_;
//public:

  // returns
  //static bool lt();
};
*/

struct Vertex
{
  IntPoint2D point;
  Vertex(): point(-1,1){}
  Vertex(const IntPoint2D& p): point(p){}
};

struct Edge
{
  float weight;
  Edge(): weight(0){}
  Edge(float w): weight(w){}
};


class ConnectedPoints
{
  // Each graph node is an occupied grid point in the voronoi map that
  // is adjacent to an unoccupied grid point. Use maps to
  // map between graph vertices and grid points.
public:
  typedef boost::adjacency_list<boost::setS, boost::vecS, boost::undirectedS,
                                Vertex, Edge> t_graph;
//  typedef boost::adjacency_list<boost::setS, boost::listS, boost::undirectedS,
//                                Vertex, Edge> t_graph;

  typedef t_graph::vertex_descriptor t_vertex;
  typedef std::map<t_vertex, t_vertex> t_v2v;
  typedef std::map<t_vertex, float> t_v2f;
  typedef std::map<t_vertex, size_t> t_v2i;


  typedef std::tr1::unordered_map<IntPoint2D, t_graph::vertex_descriptor> t_p2v;
//  typedef std::map<t_graph::vertex_descriptor, IntPoint2D> t_v2p;

private:

  float threshold_;

  t_p2v p2v_;
//  t_v2p v2p_;
  t_graph graph_;

  t_vertex getVertex(const IntPoint2D& p);

  // Conditional add and delete from the graph
  void condRemoveVertex(const IntPoint2D& p);
  t_vertex condAddVertex(const IntPoint2D& p);

  template<typename Map>
  bool hasFreeAdjacent(const Map& map, const IntPoint2D& point) const
  {
    int maxX = map.shape()[0] -1;
    int maxY = map.shape()[1] -1;

    int x; int y;
    for (int dx = -1; dx <= 1; ++dx)
    {
      x = dx + point.x();
      if (x < 0 || x > maxX) continue;
      for (int dy = -1; dy <= 1; ++dy)
      {
        if (dx == 0 && dy == 0) continue;
        y = dy + point.y();
        if (y < 0 || y > maxY) continue;
        if (map[x][y].isFree())return true;
      }
    }
    return false;
  }


  template<typename Map>
  void updateAdjacentEdges(const Map& map, const IntPoint2D& point)
  {
    int maxX = map.shape()[0] -1;
    int maxY = map.shape()[1] -1;
    static float sqrt2 = std::sqrt(2);

    t_p2v::iterator it = p2v_.find(IntPoint2D(point.x(),point.y()));
    BOOST_ASSERT(it != p2v_.end());
    t_vertex pv = it->second;

    int x; int y;
    for (int dx = -1; dx <= 1; ++dx)
    {
      x = dx + point.x();
      if (x < 0 || x > maxX) continue;
      for (int dy = -1; dy <= 1; ++dy)
      {
        if (dx == 0 && dy == 0) continue;
        y = dy + point.y();
        if (y < 0 || y > maxY) continue;
        t_p2v::iterator it = p2v_.find(IntPoint2D(x,y));
        if (it == p2v_.end()) continue;
        t_vertex apv= it->second;

        if (dx == 0 || dy == 0)
          boost::add_edge(pv, apv, Edge(1.0), graph_);
        else
          boost::add_edge(pv, apv, Edge(sqrt2), graph_);
      }
    }
  }


  void buildVertexIndexMap(t_v2i& v2i);

  // To be rebuilt after each update of the points
  void buildPoint2VertexMap();

public:
  ConnectedPoints(): threshold_(0.0){}
  ConnectedPoints(float threshold): threshold_(threshold){}

  void setThreshold(float threshold)
  {
    threshold_ = threshold;
  }

  template<typename Map, typename It>
  void updatePoints(const Map& map, It begin, It end)
  {
    // If I use a graph with boost::vecS for vertices then it
    // looks like I have to rebuild the graph on any removal
    // Otherwise removing a list of vertices will be difficult.

    std::list<IntPoint2D> toadd;
    std::list<IntPoint2D> toremove;

    while (begin != end)
    {
      const IntPoint2D& p = *begin++;
      // If the point on the grid is now free then delete any
      // corresponding vertex in the graph.
      if (map[p.x()][p.y()].isFree())
      {
        toremove.push_back(p);
      }
      // If the point on the grid is now occupied; then check
      // if is adjacent to a free point. If it is adjacent to
      // a free point then make sure there is a corresponding
      // vertex in the graph. If not make sure there is no
      // corresponding vertex.
      else
      {
        if (hasFreeAdjacent(map, p))
          toadd.push_back(p);
        else
          toremove.push_back(p);
      }
    }
    BOOST_FOREACH(const IntPoint2D& p, toremove)
    {
      condRemoveVertex(p);
    }
    buildPoint2VertexMap();
    BOOST_FOREACH(const IntPoint2D& p, toadd)
    {
      condAddVertex(p);
    }
    BOOST_FOREACH(const IntPoint2D& p, toadd)
    {
      updateAdjacentEdges(map, p);
    }
  }

  bool isBoundedlyConnected(const IntPoint2D& a, const IntPoint2D& b);

};

#endif /* CONNECTED_POINTS_H_ */
