#ifndef GRID2D_H
#define GRID2D_H

// This class is heavily focused towards integration with the nav_msgs::OccupancyGrid class.
// NOTE: I need to refactor this class to make it an abstract creator class for creating multi_array objects
// that are 2D grids, and not simply encapsulate a multi_array within another class. Then all the
// update and transform functions should simply be template functions for operating on
// certain types of multi_array objects.

#include <stdexcept>
#include <iostream>
#include <memory>
#include <utility>
#include <limits>
#include <algorithm>
#include <tr1/unordered_set>
#include <boost/multi_array.hpp>
#include <boost/foreach.hpp>
#include "Exceptions.h"
#include "mapmetadata_tools.h"


template<class Grid>
std::auto_ptr<Grid>
create_grid2d(unsigned int width, unsigned int height)
{
  std::auto_ptr<Grid> ng(new Grid(boost::extents[width][height]));
  return ng;
}

/*

template<class Grid>
std::auto_ptr<Grid>
create_grid2d(const Grid& other)
{
  std::auto_ptr<Grid> ng(new Grid(other));
  return ng;
}

template<class ThisGrid, class ThatGrid>
std::auto_ptr<ThisGrid>
create_grid2d(const ThatGrid& other)
{
  unsigned int xsize = other.shape()[0];
  unsigned int ysize = other.shape()[1];

  std::auto_ptr<ThisGrid> ng(new ThisGrid(boost::extents[xsize][ysize]));
  return ng;
}
*/

template<class ThisGrid, class ThatGrid>
std::auto_ptr<ThisGrid>
create_grid2d(const ThatGrid& other)
{
  std::auto_ptr<ThisGrid> ng(new ThisGrid(other));
  return ng;
}

template<class ThisGrid, class ThatGrid, class Fn>
std::auto_ptr<ThisGrid>
create_grid2d(const ThatGrid& other, Fn fn)
{
  unsigned int xsize = other.shape()[0];
  unsigned int ysize = other.shape()[1];

  std::auto_ptr<ThisGrid> ng(new ThisGrid(boost::extents[xsize][ysize]));

  for (unsigned int x =0; x < xsize; ++x)
  {
    for (unsigned int y=0; y < ysize; ++y)
    {
      (*ng)[x][y] = fn(other[x][y]);
    }
  }
  return ng;
}


template<typename G, typename F>
void for_each_adjacent(const G& grid, const IntPoint2D& point, F fn)
{
  int maxx = grid.shape()[0]-1;
  int maxy = grid.shape()[1]-1;

  int x; int y;
  for (int dx = -1; dx <= 1; ++dx)
  {
    x = dx + point.x();
    if (x < 0 || x > maxx) continue;
    for (int dy = -1; dy <= 1; ++dy)
    {
      if (dx == 0 && dy == 0) continue;
      y = dy + point.y();
      if (y < 0 || y > maxy) continue;
      fn(point, IntPoint2D(x,y));
    }
  }
}


template<class Grid>
void adjacents24(const Grid& grid, const IntPoint2D& point, std::vector<IntPoint2D>& adjlist)
{
  int maxx = grid.shape()[0]-1;
  int maxy = grid.shape()[1]-1;

  int x; int y;
  for (int dx = -2; dx <= 2; ++dx)
  {
    x = dx + point.x();
    if (x < 0 || x > maxx) continue;
    for (int dy = -2; dy <= 2; ++dy)
    {
      if (dx == 0 && dy == 0) continue;
      y = dy + point.y();
      if (y < 0 || y > maxy) continue;
      adjlist.push_back(IntPoint2D((unsigned int)x, (unsigned int)y));
    }
  }
}

template<class Grid>
void adjacents8(const Grid& grid, const IntPoint2D& point, std::vector<IntPoint2D>& adjlist)
{
  int maxx = grid.shape()[0]-1;
  int maxy = grid.shape()[1]-1;

  int x; int y;
  for (int dx = -1; dx <= 1; ++dx)
  {
    x = dx + point.x();
    if (x < 0 || x > maxx) continue;
    for (int dy = -1; dy <= 1; ++dy)
    {
      if (dx == 0 && dy == 0) continue;
      y = dy + point.y();
      if (y < 0 || y > maxy) continue;
      adjlist.push_back(IntPoint2D((unsigned int)x, (unsigned int)y));
    }
  }
}


template<class Grid>
void adjacents4(const Grid& grid, const IntPoint2D& point, std::vector<IntPoint2D>& adjlist)
{
  int maxx = grid.shape()[0]-1;
  int maxy = grid.shape()[1]-1;

  int x; int y;
  for (int dx = -1; dx <= 1; ++dx)
  {
    x = dx + point.x();
    if (x < 0 || x > maxx) continue;
    for (int dy = -1; dy <= 1; ++dy)
    {
      if (dx == 0 && dy == 0) continue;
      y = dy + point.y();
      if (y < 0 || y > maxy) continue;
      if (dx != 0 && dy != 0) continue;
      adjlist.push_back(IntPoint2D((unsigned int)x, (unsigned int)y));
    }
  }
}



template<class C>
inline void check_grid2d(const C& o)
{
  if (o.num_dimensions() != 2)
  {
    BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("Not a 2 dimensional multi-array")));
  }
}




template<class C1, class C2>
inline void check_matching_shape(const C1& o1, const C2& o2)
{
  if (o1.num_dimensions() != o2.num_dimensions())
  {
    BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("unmatched dimensions of input and output sources")));
  }
  for (unsigned int i = 0; i < o1.num_dimensions(); ++i)
  {
    if (o1.shape()[i] != o2.shape()[i])
    {
      BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("unmatched shape of input and output sources")));
    }
  }
}

// returns an array of all the points in the grid
template<class C>
void all_points(C& grid, std::vector<IntPoint2D>& points)
{
  check_grid2d(grid);
  unsigned int xsize = grid.shape()[0];
  unsigned int ysize = grid.shape()[1];

  for (unsigned int x = 0; x < xsize; ++x)
  {
    for (unsigned int y = 0; y < ysize; ++y)
    {
      points.push_back(IntPoint2D(x, y));
    }
  }
}

// returns an array of all the points in the grid
template<class C>
void all_points(C& grid, std::tr1::unordered_set<IntPoint2D>& points)
{
  check_grid2d(grid);
  unsigned int xsize = grid.shape()[0];
  unsigned int ysize = grid.shape()[1];

  for (unsigned int x = 0; x < xsize; ++x)
  {
    for (unsigned int y = 0; y < ysize; ++y)
    {
      points.insert(IntPoint2D(x, y));
    }
  }
}


template<class VT>
inline const VT& pass_through_function(const VT& v){ return v; }

template<class VT>
struct PassThroughFunctor{
inline const VT& operator()(const VT& v){ return v; }
};

// NOTE: In C++-0x template functions will be allowed default template classes so hopefully the following will work.
//template<class InMA, class OutMA, class Fn=PassThroughFunctor<typename InMA::element >

template<class InMA, class OutMA, class Fn>
void grid2d_update(const InMA& in_ma, OutMA& out_ma, std::vector<IntPoint2D>& out_points, Fn fn)
{
  check_grid2d(in_ma);
  check_grid2d(out_ma);
  check_matching_shape(in_ma, out_ma);
  unsigned int xsize = in_ma.shape()[0];
  unsigned int ysize = in_ma.shape()[1];

  for (unsigned int x = 0; x < xsize; ++x)
  {
    for (unsigned int y = 0; y < ysize; ++y)
    {
      if (out_ma[x][y] != fn(in_ma[x][y]))
      {
        out_ma[x][y] = fn(in_ma[x][y]);
        out_points.push_back(IntPoint2D(x, y));
      }
    }
  }
}

template<class InMA, class OutMA, class Fn>
void grid2d_update(const InMA& in_ma, const std::vector<IntPoint2D>& in_points,
            OutMA& out_ma, std::vector<IntPoint2D>& out_points, Fn fn)
{
  check_grid2d(in_ma);
  check_grid2d(out_ma);
  check_matching_shape(in_ma, out_ma);

  BOOST_FOREACH( const IntPoint2D& point, in_points)
  {
    unsigned int x = point.x();
    unsigned int y = point.y();
    if (out_ma[x][y] != fn(in_ma[x][y]))
    {
      out_ma[x][y] = fn(in_ma[x][y]);
      out_points.push_back(point);
    }
  }
}

template<class InMA, class OutMA, class Fn>
void grid2d_update(const InMA& in_ma, const std::vector<IntPoint2D>& in_points,
            OutMA& out_ma, Fn fn)
{
  check_grid2d(in_ma);
  check_grid2d(out_ma);
  check_matching_shape(in_ma, out_ma);

  BOOST_FOREACH( const IntPoint2D& point, in_points)
  {
    unsigned int x = point.x();
    unsigned int y = point.y();
    if (out_ma[x][y] != fn(in_ma[x][y]))
    {
      out_ma[x][y] = fn(in_ma[x][y]);
    }
  }
}

template <class T>
void print(const T& grid, std::size_t skip)
{
  typedef boost::multi_array_types::index_range t_range;
  typedef typename T::template const_array_view<2>::type t_const_view;

  t_const_view myview =
        grid[boost::indices[t_range(0, grid.shape()[0], skip)]
                               [t_range(0, grid.shape()[1], skip)]];

  for (int y = myview.shape()[1]-1; y >= 0; --y)
  //for (unsigned int y = 0; y < myview.shape()[1]; ++y)
  {

    for (unsigned int x = 0; x < myview.shape()[0]; ++x)
    {
      typename T::element tmp = myview[x][y];
      if ((signed char) tmp < 0) std::cout << ".";
      else if ((signed char) tmp == 0) std::cout << " ";
      else
      {
        unsigned int val = (unsigned int) tmp / 11;
        std::cout << val;
      }
    }
    std::cout << std::endl;
  }
}


// Copy a smaller grid to some area of a larger grid. Relies on overloading the
// assignment operator (operator=) of the underlying class.
template<class Grid>
void grid2d_copy_to(const Grid& input, Grid& output, const IntPoint2D& new_origin)
{

  // Make sure the input + shift points fit in the output grid
  int out_xsize = output.shape()[0];
  int out_ysize = output.shape()[1];

  int in_new_maxx = new_origin.x() + input.shape()[0] - 1;
  int in_new_maxy = new_origin.y() + input.shape()[1] - 1;

  for (int out_x = 0; out_x < out_xsize; ++out_x)
  {
    for (int out_y = 0; out_y < out_ysize; ++out_y)
    {
      if ((out_x >= new_origin.x() && out_x <= in_new_maxx) &&
          (out_y >= new_origin.y() && out_y <= in_new_maxy))
      {
        int in_x = out_x - new_origin.x();
        int in_y = out_y - new_origin.y();
        output[out_x][out_y] = input[in_x][in_y];
      }
      else if (out_x > in_new_maxx && out_y > in_new_maxy)
      {
        return;
      }
    }
  }
}

// Same as above but the user specifies a default value for those cells
// that aren't copied.
// Copy a smaller grid to some area of a larger grid. Relies on overloading the
// assignment operator (operator=) of the underlying class.
template<class Grid, class Cell>
void grid2d_copy_to(const Grid& input, Grid& output, const Cell& defvalue, const IntPoint2D& new_origin)
{

  // Make sure the input + shift points fit in the output grid
  int out_xsize = output.shape()[0];
  int out_ysize = output.shape()[1];

  int in_new_maxx = new_origin.x() + input.shape()[0] - 1;
  int in_new_maxy = new_origin.y() + input.shape()[1] - 1;

  for (int out_x = 0; out_x < out_xsize; ++out_x)
  {
    for (int out_y = 0; out_y < out_ysize; ++out_y)
    {
      if ((out_x >= new_origin.x() && out_x <= in_new_maxx) &&
          (out_y >= new_origin.y() && out_y <= in_new_maxy))
      {
        int in_x = out_x - new_origin.x();
        int in_y = out_y - new_origin.y();
        output[out_x][out_y] = input[in_x][in_y];
      }
      else
      {
        output[out_x][out_y] = defvalue;
      }
    }
  }
}




// Shift values across
template<class C>
void shift(C& grid, const IntPoint2D& init_origin,
           const IntPoint2D& shift_origin, const IntSize2D& shift_size)
{
  if (init_origin == shift_origin)return;
  check_grid2d(grid);
  unsigned int xsize = grid.shape()[0];
  unsigned int ysize = grid.shape()[1];

  if ((init_origin.x() + shift_size.x() > xsize) || (init_origin.y() + shift_size.y() > ysize) ||
      (shift_origin.x() + shift_size.x() > xsize) || (shift_origin.y() + shift_size.y() > ysize))
  {
    BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("Invalid shifting of grid")));
  }

  // Create the new grid and swap the old grid elements to the new
  // for x(y)_init_origin < x(y)_shift_origin then swap in descending order otherwise use ascending
  // order. This will prevent overwriting.
  int xshift = shift_origin.x() - init_origin.x();
  int yshift = shift_origin.y() - init_origin.y();

  if (init_origin.x() < shift_origin.x())
  {
    if (init_origin.y() < shift_origin.y())
    {
      for (int x = init_origin.x()+shift_size.x()-1; x >= (int)init_origin.x(); --x)
        for (int y = init_origin.y()+shift_size.y()-1; y >= (int)init_origin.y(); --y)
          std::swap(grid[x][y], grid[x+xshift][y+yshift]);
    }
    else
    {
      for (int x = init_origin.x()+shift_size.x()-1; x >= (int)init_origin.x(); --x)
        for (int y = init_origin.y(); y <= (int)(init_origin.y()+shift_size.y()-1); ++y)
          std::swap(grid[x][y], grid[x+xshift][y+yshift]);
    }
  }
  else
  {
    if (init_origin.y() < shift_origin.y())
    {
      for (int x = init_origin.x(); x <= (int)(init_origin.x()+shift_size.x()-1); ++x)
        for (int y = init_origin.y()+shift_size.y()-1; y >= (int)init_origin.y(); --y)
          std::swap(grid[x][y], grid[x+xshift][y+yshift]);
    }
    else
    {
      for (int x = init_origin.x(); x <= (int)(init_origin.x()+shift_size.x()-1); ++x)
        for (int y = init_origin.y(); y <= (int)(init_origin.y()+shift_size.y()-1); ++y)
          std::swap(grid[x][y], grid[x+xshift][y+yshift]);
    }
  }
}

// Some standard instantiations of the grid.
typedef boost::multi_array<signed char, 2> t_occgrid;
typedef boost::const_multi_array_ref<signed char, 2> t_const_occgrid_ref;

template<typename GC>
GC& get_cell(boost::multi_array<GC, 2>& grid, const IntPoint2D& point)
{
  if (point.x() < 0 || point.x() >= (int) grid.shape()[0] ||
      point.y() < 0 || point.y() >= (int) grid.shape()[1])
  {
    BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("The point is out of range of the grid")));
  }
  return grid[point.x()][point.y()];
}

template<typename GC>
const GC& get_cell(const boost::multi_array<GC, 2>& grid, const IntPoint2D& point)
{
  if (point.x() < 0 || point.x() >= (int) grid.shape()[0] ||
      point.y() < 0 || point.y() >= (int) grid.shape()[1])
  {
    BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("The point is out of range of the grid")));
  }
  return grid[point.x()][point.y()];
}

#endif // GRID2D_H
