/*
 * map.h
 *
 * Implements a cleaned occupancy grid determined by a threshold over the
 * probabilities. Cells are divided into unknown, free, and occupied.
 * Unknown cells can be treated as free or occupied depending on the following:
 * - Assume  occupied if the cell is unknown and has a path of unknown
 *   adjacents to an occupied cell.
 * - Assume free if the cell is unknown and there is a path of unknown
 *   adjacent to a free cell, and there is no path to an occupied cell.
 * A bushfire algorithm is used to work out the assumed cells.
 *
 *  Created on: 16/08/2011
 *      Author: daver
 */

#ifndef MAP_H_
#define MAP_H_

#include <tr1/functional>
#include <boost/functional/hash.hpp>
#include <boost/multi_array.hpp>
#include "mapmetadata_tools.h"
#include "bushfire2.h"
#include "grid2d.h"

struct OccMapCell
{
  typedef enum {unknown=-1, free=0, occupied=1 } t_occupancy;
  typedef enum { assume_none=-2, assume_occupied=-3, assume_free=-4 } t_unknown;

  t_occupancy occupancy_;
  t_unknown unknown_;

  IntPoint2D ancestor_; // for an unknown cell the cell that gives it its value
  int dist_; // distance to the ultimate ancestor

  bool isFree() const
  {
    return (occupancy_ == free ||
            (occupancy_ == unknown && unknown_ == assume_free));
  }

  bool isOccupied() const
  {
    return (occupancy_ == occupied ||
            (occupancy_ == unknown && unknown_ == assume_occupied));
  }

  bool isUnknown() const
  {
    return (occupancy_ == unknown && unknown_ == assume_none);
  }

  void setUnknown()
  {
    const limits& lim = limits::instance();
    occupancy_ = unknown;
    unknown_ = assume_none;
    ancestor_ = lim.c_invalid_point;
    dist_ = 0;
  }

  void setFree(const IntPoint2D& p)
  {
    occupancy_ = free;
    unknown_ = assume_none;
    ancestor_ = p;
    dist_ = 0;
  }

  void setOccupied(const IntPoint2D& p)
  {
    occupancy_ = occupied;
    unknown_ = assume_none;
    ancestor_ = p;
    dist_ = 0;
  }

  OccMapCell(): occupancy_(unknown), unknown_(assume_none)
  {
    const limits& lim = limits::instance();
    ancestor_ = lim.c_invalid_point;
    dist_ = 0;
  }
};

typedef boost::multi_array<OccMapCell, 2> t_occmap;

void occmap_copy_to(const t_occmap& input, t_occmap& output,
                    const IntPoint2D& new_origin, std::vector<IntPoint2D>& newpoints);

class OccMapUpdater
{
  // Convert occupancy grid values to
  struct OccGridThresholdFunc
  {
    int threshold_;
    inline int operator()(int val)
    {
      if (val < 0){ return -1; }
      else if (val >= threshold_){ return 100; }
      return 0;
    }
    OccGridThresholdFunc(int threshold): threshold_(threshold){}
  } thresholdfn_;

  // Two waves - first wave clears existing unknowns
  //             second wave recalculates
  typedef BushFire2<int, IntPoint2D> t_bushfire;
  t_bushfire propagateOccupied_;
  t_bushfire propagateFree_;
  t_bushfire propagateUnknown_;

  t_occmap* occmap_;
  std::vector<IntPoint2D> adjacents_; // make it a member so we don't have to keep reallocating
  t_pointset changes_;

  t_pointset pu_visited_;

  void runFillOccupied(t_pointset& doneset);
  void runFillFree(t_pointset& doneset);

public:
  OccMapUpdater(int threshold = 50);

  void clear(t_bushfire& bf, const int& val, const IntPoint2D& p);
  void fillOccupied(t_bushfire& bf, const int& val, const IntPoint2D& p);
  void fillFree(t_bushfire& bf, const int& val, const IntPoint2D& p);

  /* Update the occupancy map based on a list of points that are
   * known to have changed.
   */
  template<typename It>
  void update(t_occmap& occmap, const t_occgrid& occgrid, It it, It end)
  {
    check_matching_shape(occgrid, occmap);
    occmap_ = &occmap;
    changes_.clear();
    propagateFree_.clear();
    propagateUnknown_.clear();
    propagateOccupied_.clear();
    pu_visited_.clear();

    std::auto_ptr<t_occmap> origoccmap = create_grid2d<t_occmap, t_occmap>(occmap);

    t_pointset visited;
    while (it != end)
    {
      IntPoint2D p = *it++;
      if (visited.find(p) != visited.end()) continue;
      visited.insert(p);
      int pogv = thresholdfn_(get_cell(occgrid, p));
      OccMapCell& pomc = get_cell(occmap, p);

      if (pogv == 100 && pomc.occupancy_ == OccMapCell::occupied) continue;
      if (pogv == 0 && pomc.occupancy_ == OccMapCell::free) continue;
      // if (pogv == -1 && pomc.occupancy_ == OccMapCell::unknown) continue;

      changes_.insert(p);
      if (pogv == -1)
      {
        pomc.setUnknown();
        propagateUnknown_.push(0,p);
      }
      if (pogv == 100)
      {
        pomc.setOccupied(p);
        propagateOccupied_.push(0,p);
      }
      if (pogv == 0)
      {
        pomc.setFree(p);
        propagateUnknown_.push(0,p);
        propagateFree_.push(0,p);
      }
    }
    propagateUnknown_.update(t_bushfire::make_callback(this, &OccMapUpdater::clear));
    propagateOccupied_.update(t_bushfire::make_callback(this, &OccMapUpdater::fillOccupied));
    propagateFree_.update(t_bushfire::make_callback(this, &OccMapUpdater::fillFree));

    t_pointset realchanges;
    BOOST_FOREACH(const IntPoint2D& p, changes_)
    {
      OccMapCell& poc = get_cell(*origoccmap, p);
      OccMapCell& pnc = get_cell(occmap, p);
      if ((poc.isFree() != pnc.isFree()) ||
          (poc.isOccupied() != pnc.isOccupied())) realchanges.insert(p);
    }
    changes_ = realchanges;
  }

  void update(t_occmap& occmap, const t_occgrid& occgrid);
  const t_pointset& changes() const { return changes_; }
};



#endif /* MAP_H_ */
