/*********************************************************************************************
 * RegionMapper is the main class for a region mapping object. The constructor sets up
 * - a map topic for the input occupancy grid.
 * - output topic for the output region map.
 *
 * The regionmapper maintains a set of maps:
 * - the basic occupancy grid (occgrid)
 * - the occupancy map (occmap) applies thresholds to occgrid: free, occuppied, unknown.
 * -
 *
 * Functions:
 * - updateOccGridCB is the callback function that is called to update the occupancy grid.
 *   In-turn it calls initMaps for the first callback and updateMaps for all
 *   subsequent callbacks.
 *
 *********************************************************************************************/



#ifndef REGIONMAPPER_H
#define REGIONMAPPER_H

#include <memory>
#include <opencv2/highgui/highgui.hpp>
#include <ros/ros.h>
#include <nav_msgs/MapMetaData.h>
#include <nav_msgs/OccupancyGrid.h>
#include "mapmetadata_tools.h"
#include "grid2d.h"
#include "map.h"
#include "basic_map.h"
#include "basic_map_updater.h"
#include "voronoi.h"
#include "voronoi_updater.h"
#include "voronoi_region_calc.h"

class RegionMapper
{
public:
  RegionMapper();
  ~RegionMapper();


  void updateOccGridCB(const nav_msgs::OccupancyGridPtr &oc);

private:
  bool gui_;

  ros::NodeHandle node_;
  ros::Subscriber map_sub_;
  ros::Publisher regions_pub_;

  // ROS parameters used by RegionMapper
  std::string map_frame_;
  std::string map_frame_id_;
  std::string map_topic_;
  std::string regions_topic_;

  // The mapmetadata with details of the size and resolution of the maps
  nav_msgs::MapMetaData mapmetadata_;

  // The different maps:
  // - OccupancyGrid: 0-100 occupancy probability
  // - OccupancyMap: 0 represents free space, 1 occupied and -1 unknown
  std::auto_ptr<t_occgrid> occgrid_;

  // Conversion functor to convert probability points on the occupancy grid
  // to a occupied/free based on a threshold (unknown remains unknown).
  std::auto_ptr<t_occmap> occmap_;
  OccMapUpdater occmapupdater_;

  // Voronoi Map
  std::auto_ptr<t_voronoimap> voronoimap_;
  // BasicMap2VoronoiMap voronoimapfilter_;
  OccMap2VoronoiMap voronoimapfilter_;
  VoronoiMapUpdater voronoimapupdater_;
  VoronoiRegionCalc voronoigrouper_;


  void initMaps(const nav_msgs::OccupancyGridPtr &oc);
  void updateMaps(const nav_msgs::OccupancyGridPtr &oc);

  void publishRegions(const VoronoiRegionMap& vrm, const nav_msgs::MapMetaData& mmd);
};

#endif // REGIONMAPPER
