/*
 * max_voronoi_calc.h
 *
 *  Created on: 11/10/2011
 *      Author: daver
 */

#ifndef GROUP_VORONOI_CALC_H_
#define GROUP_VORONOI_CALC_H_

#include <boost/shared_ptr.hpp>
#include "mapmetadata_tools.h"
#include "bushfire2.h"
#include "voronoi.h"
#include "math.h"

class VoronoiRegion
{
public:
    IntPoint2D midpoint_;
    unsigned int dist_;
    t_vpointmap vpoints_;  // voronoi points in the region
    t_pointset ccpoints_; // critical corridor points bordering the region

    std::vector<IntPoint2D> polygon_;
    unsigned long id_; // Id is assigned when the polygon is created

    static unsigned long sm_nextid_;

    VoronoiRegion(const IntPoint2D& p, double d);
    VoronoiRegion(const VoronoiRegion& r);

    bool isValid() const;
};

/* Matches voronoi points to a region. A voronoi point can belong to
 * only one region.  */

class VoronoiRegionMap
{
public:
    typedef boost::shared_ptr<VoronoiRegion> t_region_sptr;
    typedef std::tr1::unordered_map<IntPoint2D, t_region_sptr> t_p2r;
    t_p2r p2r_;   // A map of voronoi points to the region that they belong to
    t_p2r mp2r_;  // Maps the mid-point (voronoi) point of a region to its region.
    t_p2r cc2r_;  // A map of critical corridor points to the regions they belong to.

    // A map of intersecting regions (by id)
//    std::multimap<unsigned long, unsigned long> intersect_;
    std::set<std::pair<unsigned long, unsigned long> >intersect_;


    VoronoiRegionMap(){}

    // helper functions for regions
    void newRegion(const IntPoint2D& vpoint, double distance);

    // a new region bounded by a critical corridor point
    void newRegion(const IntPoint2D& point, double distance,
                   const IntPoint2D& ccpoint);

    // remove the region to which the point belongs
    // FIXUP. DPR20111023
    // NOTE: Currently I only track the voronoi points in a region. This leads
    // to a potential problem where the region itself may change but its
    // voronoi points might not. This would not be picked up by the current
    // system and the region would not be rebuilt. While it is a bug and
    // maybe should be fixed I think this would be very unlikely to happen.
    bool removeRegion(const IntPoint2D& point);


    // Remove the point from any associated region but do not delete the region.
    void removePoint(const IntPoint2D& point);


    void addPoint(const IntPoint2D& regionpoint,
                  const IntPoint2D& newpoint, double vdist,
                  const IntPoint2D& newendpoint, unsigned int distance);
    void mergeRegions(const IntPoint2D& a, const IntPoint2D& b);
    bool inRegion(const IntPoint2D& point);
    bool inRegionBoundary(const IntPoint2D& point);

    VoronoiRegion& region(const IntPoint2D& point) { return *(p2r_[point]); }
    t_pointset points(const IntPoint2D& point)
        {
            return keys(p2r_[point]->vpoints_);
        }
    bool sameRegion(const IntPoint2D& a, const IntPoint2D& b);
    std::auto_ptr<VoronoiRegionMap> clone() const; // Performs a deep copy

    std::pair<t_p2r::const_iterator, t_p2r::const_iterator> regions() const;
    size_t numRegions() const;
};

// Group voronoi points based on the critical corridors. This is used to
// group a set of voronoi points as belonging to a region. From this we can get
// the bounding area of the region; shrink it to fit the
class VoronoiRegionCalc
{
    typedef BushFire2<int, IntPoint2D> t_bushfire2;
    t_pointset changedpoints_;
    t_bushfire2 propagate_;
    t_voronoimap* gridptr_;
    std::vector<IntPoint2D> adjacents_;
    VoronoiRegionMap regionmap_;

    struct EndPoint
    {
        IntPoint2D p_; // point
        unsigned int d_; // distance
        EndPoint(){}
        EndPoint(const IntPoint2D& p, unsigned int d): p_(p), d_(d){}
    };
    typedef std::tr1::unordered_map<IntPoint2D, EndPoint> t_p2nep;
    t_p2nep p2nep_; // map of every region point to its nearest end point


    t_voronoimap& grid();

    // Relies on the earlier pruning of the diagram
    bool isEndPoint(const IntPoint2D& p);

    // Only end voronoi points and critical corridors at the start of propagation
    void setupPropagator();

    void calculate(t_voronoimap& grid); // run the propagator
    void buildBoundingPolygons();  // build the bounding polygons for any new regions


    void calculate_intersecting_regions();

public:
    VoronoiRegionCalc();

    // Callback function to find the voronoi points within a region
    // Not to be called by an external program.
    void process(t_bushfire2& bf, const int& distance, const IntPoint2D& point);

    const VoronoiRegionMap& regionmap() const;

    template<typename It>
    void update(t_voronoimap& grid, It it, It end)
    {
        while (it != end) changedpoints_.insert(*it++);
        calculate(grid);
        buildBoundingPolygons();
        calculate_intersecting_regions();
    }

    void update(t_voronoimap& grid);
};


#endif /* GROUP_VORONOI_CALC_H_ */
