/*
 * voronoi_updater.h
 *
 *  Created on: 13/10/2011
 *      Author: daver
 */

#ifndef VORONOI_UPDATER_H_
#define VORONOI_UPDATER_H_

#include <ros/ros.h>
#include "voronoi.h"

struct VoronoiMapWave
{
  // Raise wave happens first then lower:
  // raise - raises the nearest distance to obstacles as the result of an obstacle being removed
  // lower - lowers the distance to the nearest obstacles as a result of an obstacle being added
  // skel -  a skeletonisation wave to compute the thin approximation of the voronoi points
  // corridor - find the voronoi corridor points (ie. where the line between the two
  //            obstacles passes through the voronoi point).
  enum Type { raise=1, lower=2, prune1=3, prune2=4, skeletonize=5,
              findCriticalCorridor=6, mergeCriticalCorridor=7};
};

namespace std
{
  namespace tr1
  {

    inline std::size_t hash_value(const VoronoiMapWave::Type& wt)
    {
      return hash<int>()(static_cast<int>(wt));
    }
    template<>
    struct hash<VoronoiMapWave::Type>
    {
      inline std::size_t operator()(const VoronoiMapWave::Type& wt) const
      {
        return hash_value(wt);
      }
    };
  };
};



class VoronoiMapUpdater
{
public:
  // Uses a two-wave brushfire process. First wave calculates unknown
  // adjacency paths to the border.
  typedef BushFire<t_voronoimap, VoronoiMapWave::Type, unsigned int> t_bushfire;
private:

  ConnectedPoints boundaries_;

  t_bushfire bushfire_;
  std::vector<IntPoint2D> adjacents_; // make it a member so we don't have to keep reallocating

  float resolution_; // the map resolution (metres per pixel).

  float skelObstDist_;
  float skelObstDistP_;

  float minObstDist_;
  int minObstDistSq_; // The minimum allowable distance between two obstacles
                      // to allow a voronoi point.

  float minCorrDist_;
  int minCorrDistSq_; // The minimum allowable distance between two obstacles
                      // to allow a voronoi corridor point.


  float epNeighDistM_;
  float epNeighDistP_;

  t_pointset changedPoints_; // Try to maintain a list of points that have been modified.

  bool skeletonize_; // turn on skeletonize approximation - make the voronoi map
                     // more robust to noise but very expensive.

  void calculateMinObstDistSq()
  {
    float tmp = minObstDist_/resolution_;
    minObstDistSq_ = (int)(tmp*tmp);
    ROS_DEBUG_STREAM("MinimumObstDistSq = " << minObstDistSq_);
  }

  void calculateMinCorrDistSq()
  {
    float tmp = minCorrDist_/resolution_;
    minCorrDistSq_ = (int)(tmp*tmp);
    ROS_DEBUG_STREAM("MinimumCorrDistSq = " << minCorrDistSq_);
  }


  void calculateEpNeighDistP()
  {
    float tmp = epNeighDistM_/resolution_;
    epNeighDistP_ = tmp;
  }

  void calculateSkelObstDistP()
  {
    skelObstDistP_ = skelObstDist_/resolution_;
    boundaries_.setThreshold(skelObstDistP_);
  }
  void voronoiAdjacencyList(t_bushfire& bf,
                            const IntPoint2D& point,
                            std::vector<IntPoint2D>& list);

public:

  void initialiser(t_bushfire& bf, const IntPoint2D& point);

  void raise(t_bushfire& bf, const VoronoiMapWave::Type& wt,
             const int& dm, const IntPoint2D& point);

  void lower(t_bushfire& bf, const VoronoiMapWave::Type& wt,
             const int& dm, const IntPoint2D& point);

  void checkVoronoi(t_bushfire& bf, const IntPoint2D& a, const IntPoint2D& b);
  void checkVoronoiCorridor(t_bushfire& bf, const IntPoint2D& point);

  // Pruning phase will guarantee the 1 pixel wide voronoi lines
  // First step in pruning is to fill P^8_3 patterns.
  void prune1(t_bushfire& bf, const VoronoiMapWave::Type& wt,
              const int& dm, const IntPoint2D& point);
  void prune2(t_bushfire& bf, const VoronoiMapWave::Type& wt,
              const int& dm, const IntPoint2D& point);
  bool patternMatch(t_bushfire& bf, const IntPoint2D& point);
  bool patternMatchP8_3(t_bushfire& bf, const IntPoint2D& point);

  // Skeletonisation to improve robustness to noise
  void skeletonize(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                   const int& dm, const IntPoint2D& point);
  bool patternMatchSkel(t_bushfire& bf, const IntPoint2D& point);

  // A more accurate method for identifying critical corridors (but more expensive)
  void checkCriticalCorridorAccurate(t_bushfire& bf, const t_pointset& boundary_points,
		                             const IntPoint2D& cpoint);
  bool isEndCorridorPointAccurate(t_bushfire& bf, const t_pointset& boundary_points,
								  const IntPoint2D& cpoint, const IntPoint2D& curr,
								  float max_distance, float distance);
  void findCriticalCorridorAccurate(t_bushfire& bf, const VoronoiMapWave::Type& wt,
								    const int& dm, const IntPoint2D& point);


  // Can we merge critical corridors that are close together
/*
  void mergeCriticalCorridor(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                             const int& dm, const IntPoint2D& point);
  void findClosestCriticalCorridors(t_bushfire& bf, const IntPoint2D& start,
                                    const IntPoint2D& curr, float travelDist,
                                    std::multimap<float, IntPoint2D> &cpoints);
*/

  /************************************************************/
  // Better determination of critical corridors (but more expensive)
  void mergeCriticalCorridor2(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                             const int& dm, const IntPoint2D& point);
  void findClosestCriticalCorridors2(t_bushfire& bf,
		                             const t_pointset& boundary_points,
			                         const IntPoint2D& start, const IntPoint2D& curr,
			                         float distance, float neigh_distance,
			                         t_pointset& corridor_points);



  // An end corridor point is a corridor point where we can move for epNeighbourDist
  // in any direction along a voronoi line without running out of voronoi points
  // or running into another corridor point. This will remove any corridor points
  // that don't meet this requirement.

  /*
  // Is the corridor a critical point
  void findCriticalCorridor(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                            const int& dm, const IntPoint2D& point);

  void checkEndCorridorPoint(t_bushfire& bf, const IntPoint2D& cpoint);
  bool isEndCorridorPoint(t_bushfire& bf, const IntPoint2D& cpoint,
                          const IntPoint2D& curr, float travelDist);

  void checkCriticalCorridor(t_bushfire& bf, const IntPoint2D& cpoint);
*/


  VoronoiMapUpdater();

  // The minimum distance between the two (equal) closest obstances to qualify as a voronoi point.
  void minObstDistance(float dist)
  {
    minObstDist_ = dist;
    calculateMinObstDistSq();
  }

  void minCorrDistance(float dist)
  {
    minCorrDist_ = dist;
    calculateMinCorrDistSq();
  }

  void epNeighbourDistance(float dist)
  {
    epNeighDistM_ = dist;
    calculateEpNeighDistP();
  }

  void skelObstDistance(float dist)
  {
    skelObstDist_ = dist;
    calculateSkelObstDistP();
  }

  void resolution(float res)
  {
    resolution_ = res;
    calculateMinObstDistSq();
    calculateMinCorrDistSq();
    calculateEpNeighDistP();
    calculateSkelObstDistP();
  }

  void setSkeletonize(bool val)
  {
    skeletonize_ = val;
  }

  template<typename It>
  void update(t_voronoimap& grid, It begin, It end)
  {
    BOOST_ASSERT(minObstDist_ >= 0);
    changedPoints_.clear();

    if (skeletonize_)
      boundaries_.updatePoints(grid, begin, end);

    bushfire_.update(grid, begin, end);
  }
  const t_pointset& changes() const { return changedPoints_; }
};


#endif /* VORONOI_UPDATER_H_ */
