/*
 * voronoi.cpp
 *
 *  Created on: 11/08/2011
 *      Author: daver
 */

#include <algorithm>
#include <map>
#include <cmath>
#include <tr1/unordered_map>
#include <boost/foreach.hpp>
#include "voronoi.h"
#include "grid2d.h"


VoronoiCell::VoronoiCell(const VoronoiCell& o)
{
  sqdist_ = o.sqdist_;
  occupancy_ = o.occupancy_;
  obst_ = o.obst_;
  obst2_ = o.obst2_;
  nearCorr_ = o.nearCorr_;
  ncdistpx_ = o.ncdistpx_;
  ccRatio_ = o.ccRatio_;
  corrVariance_ = o.corrVariance_;
}

VoronoiCell::VoronoiCell(VoronoiCell::t_occupancy occupancy) : occupancy_(occupancy)
{
  const limits& lim = limits::instance();
  obst_ = lim.c_invalid_point;
  obst2_ = lim.c_invalid_point;
  sqdist_ = lim.c_invalid_int_distance;
  nearCorr_ = lim.c_invalid_point;
  ncdistpx_ = lim.c_invalid_int_distance;
  ccRatio_ = 0.0;
  corrVariance_ = 0.0;
}

void VoronoiCell::operator=(const VoronoiCell::t_occupancy& o)
{
  const limits& lim = limits::instance();
  occupancy_ = o;
  obst_ = lim.c_invalid_point;
  obst2_ = lim.c_invalid_point;
  sqdist_ = lim.c_invalid_int_distance;
  nearCorr_ = lim.c_invalid_point;
  ncdistpx_ = lim.c_invalid_int_distance;
  ccRatio_ = 0.0;
  corrVariance_ = 0.0;
}

void VoronoiCell::operator=(const VoronoiCell& o)
{
  occupancy_ = o.occupancy_;
  obst_ = o.obst_;
  obst2_ = o.obst2_;
  sqdist_ = o.sqdist_;
  nearCorr_ = o.nearCorr_;
  ncdistpx_ = o.ncdistpx_;
  ccRatio_ = o.ccRatio_;
  corrVariance_ = o.corrVariance_;
}

void
VoronoiCell::validate(int minobstdistsq) const
{
  const limits& lim = limits::instance();

  if (obst_ != lim.c_invalid_point)
    {
      BOOST_ASSERT(obst2_ != lim.c_invalid_point);
      int obstdistsq = obst_.sqdistanceTo(obst2_);
      if (occupancy_ >= VoronoiCell::voronoi)
        {
          if (obstdistsq < minobstdistsq)
            {
              std::cout << "Obst1 = " << obst_ << ", Obst2 = " << obst2_
                  << std::endl;
            }
          BOOST_ASSERT(obstdistsq >= minobstdistsq);
        }
    }

}

std::auto_ptr<t_voronoimap>
voronoimap_create(const t_basicmap& input, t_pointset& boundarypoints)
{
  const limits& lim = limits::instance();
  int xsize = input.shape()[0];
  int ysize = input.shape()[1];

  std::vector<IntPoint2D> adjlist;

  std::auto_ptr<t_voronoimap> outptr(new t_voronoimap(
      boost::extents[xsize][ysize]));
  t_voronoimap& output = *outptr;

  for (int x = 0; x < xsize; ++x)
    {
      for (int y = 0; y < ysize; ++y)
        {
          IntPoint2D point(x, y);
          adjlist.clear();
          adjacents8(input, point, adjlist);

          if (input[x][y].occupancy_ == BasicMapCell::occupied)
            {
              output[x][y].occupancy_ = VoronoiCell::occupied;
              output[x][y].obst_ = point;
              output[x][y].obst2_ = point;
              output[x][y].sqdist_ = 0;
              output[x][y].nearCorr_ = lim.c_invalid_point;
              output[x][y].ccRatio_ = 0.0;

              // If it is adjacent to a free cell than add to the boundary list
              BOOST_FOREACH(const IntPoint2D& ap, adjlist)
{              if (input[ap.x()][ap.y()].occupancy_ == BasicMapCell::free)
                {
                  boundarypoints.insert(point);
                  continue;
                }
            }
        }
      else
        {
          output[x][y].occupancy_ = VoronoiCell::free;
          output[x][y].obst_ = lim.c_invalid_point;
          output[x][y].obst2_ = lim.c_invalid_point;
          output[x][y].nearCorr_ = lim.c_invalid_point;
          output[x][y].sqdist_ = lim.c_invalid_int_distance;
          output[x][y].ccRatio_ = 0.0;
        }
    }
}
return outptr;
}

std::auto_ptr<t_voronoimap>
voronoimap_create(const t_occmap& input, t_pointset& boundarypoints)
{
  const limits& lim = limits::instance();
  int xsize = input.shape()[0];
  int ysize = input.shape()[1];

  std::vector<IntPoint2D> adjlist;

  std::auto_ptr<t_voronoimap> outptr(new t_voronoimap(
      boost::extents[xsize][ysize]));
  t_voronoimap& output = *outptr;

  for (int x = 0; x < xsize; ++x)
  {
    for (int y = 0; y < ysize; ++y)
    {
      IntPoint2D point(x, y);
      adjlist.clear();
      adjacents8(input, point, adjlist);

      if (input[x][y].isOccupied())
      {
        output[x][y].occupancy_ = VoronoiCell::occupied;
        output[x][y].obst_ = point;
        output[x][y].obst2_ = point;
        output[x][y].sqdist_ = 0;
        output[x][y].nearCorr_ = lim.c_invalid_point;
        output[x][y].ccRatio_ = 0.0;

        // If it is adjacent to a free cell than add to the boundary list
        BOOST_FOREACH(const IntPoint2D& ap, adjlist)
        {
          if (input[ap.x()][ap.y()].isFree())
          {
            boundarypoints.insert(point);
            continue;
          }
        }
      }
      else
      {
        output[x][y].occupancy_ = VoronoiCell::free;
        output[x][y].obst_ = lim.c_invalid_point;
        output[x][y].obst2_ = lim.c_invalid_point;
        output[x][y].nearCorr_ = lim.c_invalid_point;
        output[x][y].sqdist_ = lim.c_invalid_int_distance;
        output[x][y].ccRatio_ = 0.0;
      }
    }
  }
  return outptr;
}


void voronoimap_validate(const t_voronoimap& vmap, int minObstDistSq)
{
  const limits& lim = limits::instance();
  int xsize = vmap.shape()[0];
  int ysize = vmap.shape()[1];

  for (int x = 0; x < xsize; ++x)
  {
    for (int y = 0; y < ysize; ++y)
    {
      IntPoint2D point(x, y);
      const VoronoiCell& vc = vmap[x][y];
      BOOST_ASSERT(vc.obst_ != lim.c_invalid_point);
      BOOST_ASSERT(vc.obst2_ != lim.c_invalid_point);
      BOOST_ASSERT(vc.sqdist_ != lim.c_invalid_int_distance);

      if (vc.occupancy_ == VoronoiCell::occupied)
      {
        BOOST_ASSERT(vc.obst_ == point);
        BOOST_ASSERT(vc.obst2_ == point);
        BOOST_ASSERT(vc.nearCorr_ == lim.c_invalid_point);
        BOOST_ASSERT(vc.sqdist_ == 0);
      }

      if (vc.occupancy_ >= VoronoiCell::free)
      {
        if (vc.occupancy_ == VoronoiCell::free) BOOST_ASSERT(vc.obst_ == vc.obst2_);
        BOOST_ASSERT(vc.sqdist_ == (int) point.sqdistanceTo(vc.obst_));
        BOOST_ASSERT(vmap[vc.obst_.x()][vc.obst_.y()].occupancy_ == VoronoiCell::occupied);
      }

      if (vc.occupancy_ >= VoronoiCell::voronoi)
      {
        int obstdistsq = vc.obst_.sqdistanceTo(vc.obst2_);
        BOOST_ASSERT(obstdistsq >= minObstDistSq);

        // NOTE: You can have a voronoi point with no nearest corridor
        // because the point many be in isolation with no connected voronoi points.
        if (vc.nearCorr_ != lim.c_invalid_point)
        {
          BOOST_ASSERT(vc.ncdistpx_ != lim.c_invalid_int_distance);
          BOOST_ASSERT(vmap[vc.nearCorr_.x()][vc.nearCorr_.y()].isCorridor());
        }
      }
    }
  }
}

void voronoimap_copy_to(const t_voronoimap& input, t_voronoimap& output,
                        const IntPoint2D& new_origin)
{
  const limits& lim = limits::instance();

  // Make sure the input + shift points fit in the output grid
  int out_xsize = output.shape()[0];
  int out_ysize = output.shape()[1];
  int in_new_maxx = new_origin.x() + input.shape()[0] - 1;
  int in_new_maxy = new_origin.y() + input.shape()[1] - 1;

  for (int out_x = 0; out_x < out_xsize; ++out_x)
    {
      for (int out_y = 0; out_y < out_ysize; ++out_y)
        {
          if ((out_x >= new_origin.x() && out_x <= in_new_maxx) && (out_y
              >= new_origin.y() && out_y <= in_new_maxy))
            {
              int in_x = out_x - new_origin.x();
              int in_y = out_y - new_origin.y();
              output[out_x][out_y] = input[in_x][in_y];
              output[out_x][out_y].obst_.shift(new_origin);
              output[out_x][out_y].obst2_.shift(new_origin);
              if (output[out_x][out_y].nearCorr_ != lim.c_invalid_point)
                output[out_x][out_y].nearCorr_.shift(new_origin);
            }
          else
            {
              output[out_x][out_y].occupancy_ = VoronoiCell::occupied;
              output[out_x][out_y].obst_ = IntPoint2D(out_x, out_y);
              output[out_x][out_y].obst2_ = IntPoint2D(out_x, out_y);
              output[out_x][out_y].sqdist_ = 0;
              output[out_x][out_y].nearCorr_ = lim.c_invalid_point;
              output[out_x][out_y].ncdistpx_ = lim.c_invalid_int_distance;
              output[out_x][out_y].ccRatio_ = 0.0;
            }
        }
    }
}

