/*
 * voronoi_updater.cpp
 *
 *  Created on: 11/08/2011
 *      Author: daver
 */

#include <algorithm>
#include <map>
#include <cmath>
#include <tr1/unordered_map>
#include <boost/foreach.hpp>
#include "voronoi.h"
#include "voronoi_updater.h"
#include "grid2d.h"
#include "max_voronoi_calc.h"
#include "voronoi_region_calc.h"
#include "bushfire2.h"

/*****************************************************************************
 * Collect a set of points by following a wall boundary point up to some
 * distance.
 * This will allow more accurate determination for end corridor points and
 * critical corridors.
 *****************************************************************************/

class FollowWallBoundary
{
	static const float c_diag_dist;
	static const float c_oth_dist;
	typedef BushFire2<float, IntPoint2D> t_bushfire;
	t_bushfire bushfire_;
	const t_voronoimap& grid_;
	t_pointset& points_;
	double max_distance_;
	std::vector<IntPoint2D> adjacents_;
public:
	FollowWallBoundary(const t_voronoimap& grid, t_pointset& points, float max_distance) :
		grid_(grid), points_(points), max_distance_(max_distance){ }

	void go(const IntPoint2D& start)
	{
		bushfire_.clear();
		bushfire_.push(0.0, start);
		bushfire_.update(t_bushfire::make_callback(this, &FollowWallBoundary::burn));
	}

	// Not to be called directly - callback for the brushfire algorithm.
	void burn(t_bushfire& bf, const float& distance, const IntPoint2D& point)
	{
		adjacents_.clear();
		adjacents8(grid_, point, adjacents_);

		// First make sure we are dealing with a boundary point.
		bool boundary = false;
		BOOST_FOREACH(const IntPoint2D&p, adjacents_)
		{
			const VoronoiCell& cell = get_cell(grid_,p);
			if (cell.isFree()){ boundary = true; break; }
		}
		if (!boundary) return;
		points_.insert(point);

		// Now add the adjacent wall points for processing
		BOOST_FOREACH(const IntPoint2D&p, adjacents_)
		{
			const VoronoiCell& cell = get_cell(grid_,p);
			if (cell.isFree()) continue;
			if (points_.count(p) == 1) continue;
			float ndist = distance + ((point.x() == p.x() || point.y() == p.y()) ? c_oth_dist : c_diag_dist);
			if (ndist > max_distance_) continue;
			bushfire_.push(ndist, p);
		}
	}
};
const float FollowWallBoundary::c_diag_dist = std::sqrt(2.0);
const float FollowWallBoundary::c_oth_dist = 1.0;



/*****************************************************************************/

VoronoiMapUpdater::VoronoiMapUpdater() :
     bushfire_(this, &VoronoiMapUpdater::initialiser), resolution_(0.0),
     minObstDistSq_(0), skeletonize_(false)
{
  bushfire_.setAction(VoronoiMapWave::raise, this, &VoronoiMapUpdater::raise);
  bushfire_.setAction(VoronoiMapWave::lower, this, &VoronoiMapUpdater::lower);
  bushfire_.setAction(VoronoiMapWave::prune1, this, &VoronoiMapUpdater::prune1);
  bushfire_.setAction(VoronoiMapWave::prune2, this, &VoronoiMapUpdater::prune2);
  bushfire_.setAction(VoronoiMapWave::skeletonize, this, &VoronoiMapUpdater::skeletonize);
  bushfire_.setAction(VoronoiMapWave::findCriticalCorridor, this,
                      &VoronoiMapUpdater::findCriticalCorridorAccurate);
  bushfire_.setAction(VoronoiMapWave::mergeCriticalCorridor, this,
                      &VoronoiMapUpdater::mergeCriticalCorridor2);
}

void VoronoiMapUpdater::voronoiAdjacencyList(t_bushfire& bf,
                                             const IntPoint2D& point,
                                             std::vector<IntPoint2D>& list)
{
  int maxX = (int) bf.maxX();
  int maxY = (int) bf.maxY();
  int x; int y;
  for (int dx = -1; dx <= 1; ++dx)
  {
    x = dx + point.x();
    if (x < 0 || x > maxX) continue;
    for (int dy = -1; dy <= 1; ++dy)
    {
      if (dx == 0 && dy == 0) continue;
      y = dy + point.y();
      if (y < 0 || y > maxY) continue;
      IntPoint2D a(x,y);
      if (!bf.cell(a).isVoronoi()) continue;
      list.push_back(a);
    }
  }
}

void VoronoiMapUpdater::initialiser(t_bushfire& bf, const IntPoint2D& point)
{
  const limits& lim = limits::instance();
  VoronoiCell& cell = bf.cell(point);
  adjacents_.clear();
  changedPoints_.insert(point); // add to the points that have been changed

  if (cell.occupancy_ == VoronoiCell::occupied)
  {
    cell.sqdist_ = 0;
    cell.obst_ = point;
    cell.obst2_ = point;
    cell.nearCorr_ = lim.c_invalid_point;
    cell.ncdistpx_ = lim.c_invalid_int_distance;
    // Only create a lower wavefront if the point is next to
    // at least one free cell.
    bf.adjacencyList(point, adjacents_);
    BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
    {
      VoronoiCell& ac = bf.cell(ap);
      if (ac.occupancy_ != VoronoiCell::occupied)
      {
        bf.queue().push(VoronoiMapWave::lower, 0, point);
        return;
      }
    }
  }
  else
  {
    cell.occupancy_ = VoronoiCell::free;
    cell.obst_ = lim.c_invalid_point;
    cell.obst2_ = lim.c_invalid_point;
    cell.sqdist_ = lim.c_invalid_int_distance;
    cell.nearCorr_ = lim.c_invalid_point;
    cell.ncdistpx_ = lim.c_invalid_int_distance;

    bf.queue().push(VoronoiMapWave::raise, 0, point);
  }
}

void VoronoiMapUpdater::raise(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                              const int& dm, const IntPoint2D& point)
{
  const limits& lim = limits::instance();
  VoronoiCell& cell = bf.cell(point);
  BOOST_ASSERT(cell.occupancy_ != VoronoiCell::occupied);

  adjacents_.clear();
  bf.adjacencyList(point, adjacents_);

  BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
  {
    VoronoiCell& ac = bf.cell(ap);
    if (ac.obst_ == lim.c_invalid_point) continue; // already set to an invalid obstacle
    IntPoint2D aop = ac.obst_;
    VoronoiCell& aoc = bf.cell(aop);

    int ap2aop_distsq = ap.sqdistanceTo(aop);
    BOOST_ASSERT(ac.sqdist_ == ap2aop_distsq);

    // Check if we've hit the limit of the raise wavefront and should start a lower wavefront
    if (ac.occupancy_ == VoronoiCell::occupied)
    {
      bf.queue().push(VoronoiMapWave::lower, 0, ap);
      continue;
    }

    else if (ac.occupancy_ == VoronoiCell::free &&
             aoc.occupancy_ == VoronoiCell::occupied &&
             ac.sqdist_ == ap2aop_distsq)
    {
      bf.queue().push(VoronoiMapWave::lower, ac.sqdist_ , ap);
      continue;
    }

    // spread the raise wavefront
    ac.occupancy_ = VoronoiCell::free;
    ac.obst_ = lim.c_invalid_point;
    ac.obst2_ = lim.c_invalid_point;
    ac.sqdist_ = lim.c_invalid_int_distance;
    ac.nearCorr_ = lim.c_invalid_point;
    ac.ncdistpx_ = lim.c_invalid_int_distance;

    bf.queue().push(VoronoiMapWave::raise, 0 , ap);
    changedPoints_.insert(ap); // add to the points that have been changed
  }
}

void VoronoiMapUpdater::lower(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                              const int& dm, const IntPoint2D& point)
{
  const limits& lim = limits::instance();
  VoronoiCell& cell = bf.cell(point);
  BOOST_ASSERT(cell.obst_ != lim.c_invalid_point && cell.obst2_ != lim.c_invalid_point);
  IntPoint2D op = cell.obst_;
  VoronoiCell& oc = bf.cell(op);
  BOOST_ASSERT(oc.occupancy_ == VoronoiCell::occupied);

  adjacents_.clear();
  bf.adjacencyList(point, adjacents_);

  BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
  {
    VoronoiCell& ac = bf.cell(ap);
    if (ac.occupancy_ == VoronoiCell::occupied) continue;
    if (ac.isVoronoi()) continue;  // Don't need to recheck
    int sqdist = ap.sqdistanceTo(op);
    if (ac.obst_ == lim.c_invalid_point ||
      (ac.obst_ != lim.c_invalid_point && sqdist < ac.sqdist_))
    {
      ac.occupancy_ = VoronoiCell::free;
      ac.obst_ = op;
      ac.obst2_ = op;
      ac.sqdist_ = sqdist;
      changedPoints_.insert(ap); // add to the points that have been changed
      bf.queue().push(VoronoiMapWave::lower, sqdist, ap);
    }
    else
    {
      checkVoronoi(bf, point, ap);
    }
  }
}

void VoronoiMapUpdater::checkVoronoi(t_bushfire& bf, const IntPoint2D& ap,
    const IntPoint2D& bp)
{
  const limits& lim = limits::instance();
  VoronoiCell& ac = bf.cell(ap);
  VoronoiCell& bc = bf.cell(bp);

  BOOST_ASSERT(ac.obst_ != lim.c_invalid_point && ac.obst2_ != lim.c_invalid_point);
  BOOST_ASSERT(bc.obst_ != lim.c_invalid_point && bc.obst2_ != lim.c_invalid_point);

  // Make sure one of ap or bp is not adjacent to its closest obstacle -
  // Improves robustness to noise (see Lau et al)
  if (ac.sqdist_ <= 2 || bc.sqdist_ <= 2) return;

  // compute distance from a to obstacle of b
  int sqd_a2obstb = ap.sqdistanceTo(bc.obst_);
  int stability_a = sqd_a2obstb - ac.sqdist_;
  if (stability_a < 0) return;

  // compute distance from b to obstacle of a
  int sqd_b2obsta = bp.sqdistanceTo(ac.obst_);
  int stability_b = sqd_b2obsta - bc.sqdist_;
  if (stability_b < 0) return;

  // compute distance between the two obstacles and make sure it
  // satisfies the minimum distance threshold.
  int sqd_obsta2obstb = ac.obst_.sqdistanceTo(bc.obst_);
  if (sqd_obsta2obstb <= 2) return;
  if (sqd_obsta2obstb < minObstDistSq_) return;

  // decide which cell to add to the voronoi diagram
  if (stability_a <= stability_b && ac.sqdist_ > 2 && !ac.isVoronoi())
  {
    changedPoints_.insert(ap); // add to the points that have been changed
    ac.occupancy_ = VoronoiCell::voronoi;
    ac.obst2_ = bc.obst_;
    ac.nearCorr_ = lim.c_invalid_point;
    ac.ncdistpx_ = lim.c_invalid_int_distance;
    checkVoronoiCorridor(bf, ap);
    bf.queue().push(VoronoiMapWave::prune1, 0, ap);
    bf.queue().push(VoronoiMapWave::prune2, 0, ap);
  }
  if (stability_b <= stability_a && bc.sqdist_ > 2 && !bc.isVoronoi())
  {
    changedPoints_.insert(bp); // add to the points that have been changed
    bc.occupancy_ = VoronoiCell::voronoi;
    bc.obst2_ = ac.obst_;
    bc.nearCorr_ = lim.c_invalid_point;
    bc.ncdistpx_ = lim.c_invalid_int_distance;
    checkVoronoiCorridor(bf, bp);
    bf.queue().push(VoronoiMapWave::prune1, 0, bp);
    bf.queue().push(VoronoiMapWave::prune2, 0, bp);
  }
}

// Prune1 - looks for P^8_3 pattern and adds the mid point to the voronoi diagram.
// Prune1 only adds voronoi points - prune2 deletes voronoi points.
void VoronoiMapUpdater::prune1(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                               const int& dm, const IntPoint2D& point)
{
  VoronoiCell& cell = bf.cell(point);
  BOOST_ASSERT(cell.isFree());
  if (dm > 0 && cell.isVoronoi()) return;
  if (dm == 0)
  {
    adjacents_.clear();
    bf.adjacencyList(point, adjacents_);
    BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
    {
      VoronoiCell& ac = bf.cell(ap);
      if (ac.isFree() && !ac.isVoronoi())
        bf.queue().push(VoronoiMapWave::prune1, 1, ap);
    }
    return;
  }
  if (!patternMatchP8_3(bf, point)) return;

  // Add to the voronoi diagram. Look at the obstacles for the
  // adjacents vonoroi points and take the two points that are
  // a minimal distance to this poitn.
  adjacents_.clear();
  bf.adjacencyList(point, adjacents_);
  t_pointset obstacles;
  BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
  {
    VoronoiCell& ac = bf.cell(ap);
    if (!ac.isVoronoi()) continue;
    obstacles.insert(ac.obst_);
    obstacles.insert(ac.obst2_);
  }
  std::multimap<int,IntPoint2D> d2op;
  BOOST_FOREACH(const IntPoint2D& op, obstacles)
  {
    d2op.insert(std::pair<int, IntPoint2D>(point.sqdistanceTo(op), op));
  }
  std::multimap<int,IntPoint2D>::iterator it = d2op.begin();
  cell.occupancy_ = VoronoiCell::voronoi;
  cell.obst_ = it->second;
  cell.sqdist_ = it->first;
  it++;
  cell.obst2_ = it->second;
  checkVoronoiCorridor(bf, point);
  bf.queue().push(VoronoiMapWave::prune2, 0, point);
  return;
}

void VoronoiMapUpdater::checkVoronoiCorridor(t_bushfire& bf, const IntPoint2D& point)
{
	VoronoiCell& cell = bf.cell(point);
	BOOST_ASSERT(cell.isVoronoi());
	// A corridor point if the distance between the two nearest obstacles is
	// equal to double the distance from the point to the nearest obstacle.
	// Which also means that the line between the two obstacles passes through
	// the voronoi point.

	float dist_between_obsts = cell.obst_.distanceTo(cell.obst2_);
	if (dist_between_obsts < std::sqrt(minCorrDistSq_)) return;
	float dist_obst1 = point.distanceTo(cell.obst_);
	float dist_obst2 = point.distanceTo(cell.obst2_);
	float min_obst = std::min(dist_obst1, dist_obst2);
	float max_obst = std::max(dist_obst1, dist_obst2);

	float var1 = (max_obst - min_obst)/min_obst;
	float var2 = ((dist_obst1+dist_obst2) - dist_between_obsts)/dist_between_obsts;


	if (var2 <= var1*0.4)
	{
		changedPoints_.insert(point);
		cell.occupancy_ = VoronoiCell::corridor;
		cell.nearCorr_ = point;
		cell.ncdistpx_ = 0;
		cell.corrVariance_ = var2;
		bf.queue().push(VoronoiMapWave::findCriticalCorridor, 0, point);
	}
/*
  int sqdist_obsts = cell.obst_.sqdistanceTo(cell.obst2_);

  if (sqdist_obsts < minCorrDistSq_) return;

  int sqdist_p2obst1 = point.sqdistanceTo(cell.obst_);
  int sqdist_p2obst2 = point.sqdistanceTo(cell.obst2_);

  int tmp1 = 3*sqdist_p2obst1 + sqdist_p2obst2;
  int tmp2 = 3*sqdist_p2obst2 + sqdist_p2obst1;

  if (sqdist_obsts >= std::min(tmp1, tmp2) && sqdist_obsts <= std::max(tmp1, tmp2))
  {
    changedPoints_.insert(point);
    cell.occupancy_ = VoronoiCell::corridor;
    cell.nearCorr_ = point;
    cell.ncdistpx_ = 0;
    bf.queue().push(VoronoiMapWave::findCriticalCorridor, 0, point);
  }
  */
}

bool VoronoiMapUpdater::patternMatchP8_3(t_bushfire& bf, const IntPoint2D& point)
{
  bool f[9];
  int i=0;
  int x = point.x();
  int y = point.y();

  for (int dx = -1; dx <= 1; ++dx) {
    int nx = x + dx;
    for (int dy = -1; dy <= 1; ++dy){
      int ny = y + dy;
      if ((nx < 0) || (nx >=  (int) bf.sizeX()) || (ny < 0) || (ny >= (int) bf.sizeY())){
        f[i] = false;
      } else {
        f[i] = bf.cell(IntPoint2D(nx,ny)).isVoronoi();
      }

      if (i == 7){
        if (f[1] && !f[3] && f[5] && f[7]){ return true; } // P^8_3
        return false;
      }
      ++i;
    }
  }
  return false;
}

void VoronoiMapUpdater::prune2(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                               const int& dm, const IntPoint2D& point)
{
  VoronoiCell& cell = bf.cell(point);
  BOOST_ASSERT(cell.isVoronoi());
  if (patternMatch(bf, point))
  {
    if (skeletonize_)
      bf.queue().push(VoronoiMapWave::skeletonize, cell.sqdist_, point);
    return;
  }

  adjacents_.clear();
  bf.adjacencyList(point, adjacents_);
  bool removeVoronoi = false;
  BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
  {
    VoronoiCell& ac = bf.cell(ap);
/*    int p2po2_dist = point.sqdistanceTo(cell.obst2_);
    int a2ao2_dist = point.sqdistanceTo(ac.obst2_);
    if (ac.isVoronoi() && (cell.sqdist_ + p2po2_dist) >= (ac.sqdist_ + a2ao2_dist))
*/
    if (ac.isVoronoi())
    {
      removeVoronoi = true;
      break;
    }
  }
  if (removeVoronoi)
  {
    cell.occupancy_ = VoronoiCell::free;
    cell.obst2_ = cell.obst_;
    return;
  }
  else
  {
    if (skeletonize_)
      bf.queue().push(VoronoiMapWave::skeletonize, cell.sqdist_, point);
  }
}

// Copied from Lau et al.
bool VoronoiMapUpdater::patternMatch(t_bushfire& bf, const IntPoint2D& point)
{
  bool f[9];
  int i=0;
  int x = point.x();
  int y = point.y();

  for (int dx = -1; dx <= 1; ++dx) {
    int nx = x + dx;
    for (int dy = -1; dy <= 1; ++dy){
      int ny = y + dy;
      if ((nx < 0) || (nx >=  (int) bf.sizeX()) || (ny < 0) || (ny >= (int) bf.sizeY())){
        f[i] = false;
      } else {
        f[i] = bf.cell(IntPoint2D(nx,ny)).isVoronoi();
      }

      if (i == 3){
        if (!f[0] && f[1] && f[3]){ return true; } // P^4_1
        if (f[0] && !f[1] && !f[3]){ return true; } // P^8_1
      }
      if (i == 5){
        if (f[1] && !f[2] && f[5]){ return true; } // P^4_1
        if (!f[1] && f[2] && !f[5]){ return true; } // P^8_1
      }
      if (i == 7){
        if (f[3] && !f[6] && f[7]){ return true; } // P^4_1
        if (!f[3] && f[6] && !f[7]){ return true; } // P^8_1
        if (!f[1] && f[3] && f[5] && !f[7]){ return true; }  // P^4_2, P^8_2
        if (f[1] && !f[3] && !f[5] && f[7]){ return true; }  // P^4_2, P^8_2
        if (f[1] && f[3] && f[5] && f[7]){ return true; } // P^8_3
      }
      ++i;
    }
  }

  if (f[5] && f[7] && !f[8]){ return true; } // P^4_1
  if (!f[5] && !f[7] && f[8]){ return true; } // P^8_1

  return false;
}

// The skeletonize process approximates the voronoi map by filtering out
// voronoi points where the obstacles are not a minimal distance apart
// along the wall. The makes the mapping much more robust and less
// susceptible to noise. The process looks at points closest to a
// obstacle first and gradually gets further away from obstacles.
// This allows us to make sure that we don't break any points on the
// voronoi diagram.
void VoronoiMapUpdater::skeletonize(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                                    const int& dm, const IntPoint2D& point)
{
  VoronoiCell& cell = bf.cell(point);
  BOOST_ASSERT(cell.isVoronoi());
  float obstdist = cell.obst_.distanceTo(cell.obst2_);

  if (breakpoint_check(point, IntPoint2D(296,235)) ||
      breakpoint_check(point, IntPoint2D(296,236)) ||
      breakpoint_check(point, IntPoint2D(296,237)) ||
      breakpoint_check(point, IntPoint2D(296,238)))
  {
    std::cout << "BPC: " << point << " - Obstacles " << cell.obst_ << " and " << cell.obst2_ << " : "<< obstdist << std::endl;
  }

  bool keepvoronoi = false;
  // If the geometric distance is greater than the allowed distance then it is guaranteed to be ok.
  if (obstdist >= skelObstDistP_) keepvoronoi = true;

  // Make sure we don't break the diagram
  if (!keepvoronoi && patternMatchSkel(bf, point)) keepvoronoi = true;

  // Now look at the distance along the boundary
  if (!keepvoronoi && !boundaries_.isBoundedlyConnected(cell.obst_, cell.obst2_))
    keepvoronoi = true;

  if (keepvoronoi)
  {
    //bf.queue().push(VoronoiMapWave::prune1, 0, point);
    //bf.queue().push(VoronoiMapWave::prune2, 0, point);
  }
  else
  {
    cell.occupancy_ = VoronoiCell::free;
    cell.obst2_ = cell.obst_;
  }
}

// Returns true if any of the patterns match. If the patterns match then we will
// keep the point as a voronoi point.
bool VoronoiMapUpdater::patternMatchSkel(t_bushfire& bf, const IntPoint2D& point)
{
  bool f[9];
  int i=0;
  int numoccadj = -1;
  int x = point.x();
  int y = point.y();

  for (int dx = -1; dx <= 1; ++dx)
  {
    int nx = x + dx;
    for (int dy = -1; dy <= 1; ++dy)
    {
      int ny = y + dy;
      if ((nx < 0) || (nx >=  (int) bf.sizeX()) || (ny < 0) || (ny >= (int) bf.sizeY()))
      {
        f[i] = false;
      }
      else
      {
        if ((f[i] = bf.cell(IntPoint2D(nx,ny)).isVoronoi())) ++numoccadj;
      }
      ++i;
    }
  }
  BOOST_ASSERT(numoccadj>=0);

  if (numoccadj == 0) return false;
  else if (numoccadj == 1) return false;
  else if (numoccadj == 2)
  {
    if (f[0] && f[1]){ return false; }
    if (f[1] && f[2]){ return false; }
    if (f[0] && f[3]){ return false; }
    if (f[3] && f[6]){ return false; }
    if (f[2] && f[5]){ return false; }
    if (f[5] && f[8]){ return false; }
    if (f[6] && f[7]){ return false; }
    if (f[7] && f[8]){ return false; }

    if (f[3] && f[1]){ return false; }
    if (f[1] && f[5]){ return false; }
    if (f[5] && f[7]){ return false; }
    if (f[7] && f[3]){ return false; }
  }
  else if (numoccadj == 3)
  {
    if (f[0] && f[1] && f[2]){ return false; }
    if (f[6] && f[7] && f[8]){ return false; }
    if (f[0] && f[3] && f[6]){ return false; }
    if (f[2] && f[5] && f[8]){ return false; }

    if (f[3] && f[0] && f[1]){ return false; }
    if (f[1] && f[2] && f[5]){ return false; }
    if (f[5] && f[8] && f[7]){ return false; }
    if (f[7] && f[6] && f[3]){ return false; }
  }
  return patternMatch(bf, point);
}


/*
// Calculates the nearest corridor for voronoi points - and also determines a
// region to which the voronoi point belongs.
void VoronoiMapUpdater::findCriticalCorridor(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                                             const int& dm, const IntPoint2D& point)
{
	VoronoiCell& cell = bf.cell(point);
	if (!cell.isCorridor()) return;

	//  BOOST_ASSERT(bf.cell(point).isCorridor());
  if (isEndCorridorPoint(bf, point, point,0))
      checkCriticalCorridor(bf, point);
  else
    cell.occupancy_ = VoronoiCell::voronoi; // reduce to a plain voronoi point
}



bool VoronoiMapUpdater::isEndCorridorPoint(t_bushfire& bf, const IntPoint2D& cpoint,
                                           const IntPoint2D& curr, float travelDist)
{
  VoronoiCell& cell = bf.cell(curr);
  BOOST_ASSERT(cell.isVoronoi());
  if (travelDist >= epNeighDistP_) return true;
  std::vector<IntPoint2D> adjacents;
  voronoiAdjacencyList(bf, curr, adjacents);

  //  adjacencyList(curr, adjacents);
  BOOST_FOREACH(const IntPoint2D& ap, adjacents)
  {
    VoronoiCell& ac = bf.cell(ap);
    if (ac.isCorridor()) continue;
    if (cpoint.distanceTo(ap) <= cpoint.distanceTo(curr)) continue; // always move away from the corridor point
    if (isEndCorridorPoint(bf,cpoint, ap, travelDist + curr.distanceTo(ap))) return true;
  }
  return false;
}
*/

void VoronoiMapUpdater::checkCriticalCorridorAccurate(t_bushfire& bf, const t_pointset& boundary_points,
		                                              const IntPoint2D& cpoint)
{
	float criticalCorridorMinRatio  = 1.30;
	VoronoiCell& cell = bf.cell(cpoint);
	BOOST_ASSERT(cell.isCorridor());

	MaxVoronoiCalc mvcalc(bf.grid(), boundary_points);
	float cp2obstDist = (cpoint.distanceTo(cell.obst_) + cpoint.distanceTo(cell.obst2_))/2.0;

	// Rather than use a fixed neighbourhood distance instead use the distance
	// of the point to it's nearest obstacle (x2).
	const IntPoint2D mvpoint = mvcalc.calculate(cpoint, cp2obstDist*2.0);
	VoronoiCell& mvcell = bf.cell(mvpoint);
	float dist_cp2mv = cpoint.distanceTo(mvpoint);

	if (dist_cp2mv/cp2obstDist < 0.3) return;

	if (dist_cp2mv < cp2obstDist)
	{
		criticalCorridorMinRatio = 1.1 + 0.13 * dist_cp2mv/cp2obstDist;
	}

	if (cpoint == IntPoint2D(325,227))
		std::cout << "CPOINT: " << cpoint << " => MV POINT: " << mvpoint
		          << ", ccr = " <<  criticalCorridorMinRatio << std::endl;


	float mv2obstDist = (mvpoint.distanceTo(mvcell.obst_) + mvpoint.distanceTo(mvcell.obst2_))/2.0;
	float ccRatio = mv2obstDist/cp2obstDist;
	cell.ccRatio_ = ccRatio;

	if (ccRatio >= criticalCorridorMinRatio)
	{
		// Initially assign it criticalCorridor status and later we will downgrade
		// the ones that are only possible candidates
		cell.occupancy_ = VoronoiCell::criticalCorridor;
		bf.queue().push(VoronoiMapWave::mergeCriticalCorridor, 0, cpoint);
	}
}


bool VoronoiMapUpdater::isEndCorridorPointAccurate(t_bushfire& bf, const t_pointset& boundary_points,
		                                           const IntPoint2D& cpoint, const IntPoint2D& curr,
		                                           float max_distance, float distance)
{
  VoronoiCell& cell = bf.cell(curr);
  BOOST_ASSERT(cell.isVoronoi());
  if (distance >= max_distance) return true;
  std::vector<IntPoint2D> adjacents;
  voronoiAdjacencyList(bf, curr, adjacents);

  BOOST_FOREACH(const IntPoint2D& ap, adjacents)
  {
    VoronoiCell& ac = bf.cell(ap);
    if (boundary_points.count(ac.obst_) == 0 || boundary_points.count(ac.obst2_) == 0) return true;
    if (ac.isCorridor()) continue;
    if (cpoint.distanceTo(ap) <= cpoint.distanceTo(curr)) continue; // always move away from the corridor point
    if (isEndCorridorPointAccurate(bf, boundary_points, cpoint, ap,
								   max_distance, distance + curr.distanceTo(ap))) return true;
  }
  return false;
}

// Note: by the time we check this for a point it may not longer be a voronoi point, as
// it may have been pruned out. So don't assert that it must be a voronoi point.
void VoronoiMapUpdater::findCriticalCorridorAccurate(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                                                     const int& dm, const IntPoint2D& point)
{
	VoronoiCell& cell = bf.cell(point);
	if (!cell.isCorridor()) return;
	float max_distance = epNeighDistP_*0.7;
	//  BOOST_ASSERT(bf.cell(point).isCorridor());
	t_pointset basis_boundary_points;
	FollowWallBoundary fb(bf.grid(), basis_boundary_points, max_distance);
	fb.go(cell.obst_);
	fb.go(cell.obst2_);

  if (isEndCorridorPointAccurate(bf, basis_boundary_points, point, point, max_distance, 0))
  {
      checkCriticalCorridorAccurate(bf, basis_boundary_points, point);
  }
  else
    cell.occupancy_ = VoronoiCell::corridor; // reduce to a plain corridor point
}

/*
void VoronoiMapUpdater::checkCriticalCorridor(t_bushfire& bf, const IntPoint2D& cpoint)
{
  const float criticalCorridorMinRatio  = 1.4;
  VoronoiCell& cell = bf.cell(cpoint);
  BOOST_ASSERT(cell.isCorridor());

  t_pointset bp;
  MaxVoronoiCalc mvcalc(bf.grid(), bp);
  IntPoint2D cpobst = bf.cell(cpoint).obst_;
  float cpdist = cpoint.distanceTo(cpobst);

  // FIXUP - rather than use a fixed neighbourhood distance instead I'm trying out
  // the distance of the point to it's nearest obstacle (x2).
  // const IntPoint2D mvpoint = mvcalc.calculate(cpoint, bf.grid(), epNeighDistP_);
  const IntPoint2D mvpoint = mvcalc.calculate(cpoint, cpdist*2.0);

  IntPoint2D mvpobst = bf.cell(mvpoint).obst_;
  float mvdist = mvpoint.distanceTo(mvpobst);
  float ccRatio = mvdist/cpdist;
  cell.ccRatio_ = ccRatio;

  //  if (((mvdist - cpdist)/cpdist) > 0.3)
  if (ccRatio >= criticalCorridorMinRatio)
  {
    // Initially assign it criticalCorridor status and later we will downgrade
    // the ones that are only possible candidates
    cell.occupancy_ = VoronoiCell::criticalCorridor;
    bf.queue().push(VoronoiMapWave::mergeCriticalCorridor, 0, cpoint);
  }
}

// Can we merge critical corridors that are close together
void VoronoiMapUpdater::mergeCriticalCorridor(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                                              const int& dm, const IntPoint2D& point)
{
  VoronoiCell& cell = bf.cell(point);
  if (!cell.isCriticalCorridor()) return;
  std::multimap<float, IntPoint2D> closestCC;
  findClosestCriticalCorridors(bf, point, point, 0, closestCC);
  if (closestCC.empty())
  {
//    bf.queue().push(VoronoiMapWave::fillRegionsBase, 0, point);
  }
  else
  {
    // Maybe need some other decision based on some property of the criticial corridors
    // but for the moment simply downgrade this point to a normal corridor point
    cell.occupancy_ = VoronoiCell::criticalCorridorCandidate;
  }
}

// Given a vorinoimap and two critical points decide if they are on the same corridor within
// some epsilon neighbourhood.
// Note: I could do something more accurate by following wall boundaries, but this will be
// relatively expensive when performed repeatedly. So I'm doing something blunter and simply
// looking at the distance between obstacle points. This should work provided that the
// neighbourhood distance is smaller than the minimum distance to define a corridor point.
bool isSameCorridor(const t_voronoimap& vmap, float epdist, const IntPoint2D& p1, const IntPoint2D& p2)
{
  if (p1.distanceTo(p2) > epdist) return false;
  const VoronoiCell& c1 = get_cell(vmap, p1);
  const VoronoiCell& c2 = get_cell(vmap, p2);

  // Pair up the obstacles
  const IntPoint2D& p1o1 = c1.obst_;
  const IntPoint2D& p1o2 = c1.obst2_;
  const IntPoint2D& p2o1 = c2.obst_;
  const IntPoint2D& p2o2 = c2.obst2_;
  float p1o1_to_p2o1 = p1o1.distanceTo(p2o1);
  float p1o1_to_p2o2 = p1o1.distanceTo(p2o2);
  float p1o2_to_p2o1 = p1o2.distanceTo(p2o1);
  float p1o2_to_p2o2 = p1o2.distanceTo(p2o2);

  float dista, distb;

  if (p1o1_to_p2o1 < p1o1_to_p2o2)
  {
    dista = p1o1_to_p2o1;
    distb = p1o2_to_p2o2;
  }
  else
  {
    dista = p1o1_to_p2o2;
    distb = p1o2_to_p2o1;
  }
  return (dista <= epdist && distb <= epdist);
}



// Find all the critical corridor points within a neigbourhood where the two basis points
// (ie obstacle points) are also in the neigbourhood of each other.
void VoronoiMapUpdater::findClosestCriticalCorridors(t_bushfire& bf, const IntPoint2D& start,
                                                     const IntPoint2D& curr, float travelDist,
                                                     std::multimap<float, IntPoint2D> &cpoints)
{
  float epNeigh = epNeighDistP_*0.75;
  if (travelDist >= epNeigh) return;
  std::vector<IntPoint2D> adjacents;
  voronoiAdjacencyList(bf, curr, adjacents);
  BOOST_FOREACH(const IntPoint2D& ap, adjacents)
  {
    VoronoiCell& ac = bf.cell(ap);
    if (start.distanceTo(ap) <= start.distanceTo(curr)) continue;
    float ntd = travelDist+curr.distanceTo(ap);
    if (ac.isCriticalCorridor() && isSameCorridor(bf.grid(), epNeigh, start, ap))
    {
      cpoints.insert(std::pair<float, IntPoint2D>(ntd, ap));
      continue;
    }
    findClosestCriticalCorridors(bf, start, ap, ntd, cpoints);
  }
}
*/

// Can we merge critical corridors that are close together
void VoronoiMapUpdater::mergeCriticalCorridor2(t_bushfire& bf, const VoronoiMapWave::Type& wt,
                                               const int& dm, const IntPoint2D& point)
{
	//  float epNeigh = epNeighDistP_*0.75;
	float epNeigh = epNeighDistP_;
	VoronoiCell& cell = bf.cell(point);
	if (!cell.isCriticalCorridor()) return;
	t_pointset closestCC;
	t_pointset basis_boundary_points;
	FollowWallBoundary fb(bf.grid(), basis_boundary_points, epNeigh);
	fb.go(cell.obst_);
	fb.go(cell.obst2_);
	t_pointset obsts; obsts.insert(cell.obst_); obsts.insert(cell.obst2_);
	findClosestCriticalCorridors2(bf, basis_boundary_points, point, point, 0, epNeigh, closestCC);
	if (!closestCC.empty())
	{

		// If one of the other candidates has a higher critical corridor ratio then downgrade
		BOOST_FOREACH(const IntPoint2D& op, closestCC)
		{
			if (point == IntPoint2D(500,244) && op == IntPoint2D(500,238))
			{
				std::cout << "CRITICAL CORRIDOR TEST 1" << point <<  std::endl;
			}
			if (op == IntPoint2D(500,244) && point == IntPoint2D(500,238))
			{
				std::cout << "CRITICAL CORRIDOR TEST 2" << point <<  std::endl;
			}

			VoronoiCell& oc = bf.cell(op);
			if ((obsts.count(oc.obst_) == 1 && obsts.count(oc.obst2_) == 1 &&
			     oc.corrVariance_ >= oc.corrVariance_) ||
				(oc.ccRatio_ >= cell.ccRatio_))
			{
				cell.occupancy_ = VoronoiCell::criticalCorridorCandidate;
				return;
			}
		}

		// Maybe need some other decision based on some property of the criticial corridors
		// but for the moment simply downgrade this point to a normal corridor point
		// cell.occupancy_ = VoronoiCell::criticalCorridorCandidate;
	}
}


// Find all the critical corridor points within a neigbourhood where the two basis points
// (ie obstacle points) are also in the neigbourhood of each other.
void VoronoiMapUpdater::findClosestCriticalCorridors2(t_bushfire& bf,
													  const t_pointset& boundary_points,
													  const IntPoint2D& start, const IntPoint2D& curr,
													  float distance, float neigh_distance,
                                                      t_pointset& corridor_points)
{
  if (distance >= neigh_distance) return;
  std::vector<IntPoint2D> adjacents;
  voronoiAdjacencyList(bf, curr, adjacents);
  BOOST_FOREACH(const IntPoint2D& ap, adjacents)
  {
    VoronoiCell& ac = bf.cell(ap);
	if (!ac.isVoronoi()) continue;
    if (start.distanceTo(ap) <= start.distanceTo(curr)) continue;
    float ntd = distance + curr.distanceTo(ap);
    if (ac.isCriticalCorridor() && boundary_points.count(ac.obst_) > 0 && boundary_points.count(ac.obst2_) > 0)
    {
      corridor_points.insert(ap);
      continue;
    }
    findClosestCriticalCorridors2(bf, boundary_points, start, ap, ntd, neigh_distance, corridor_points);
  }
}


