#!/usr/bin/env python

#--------------------------------------------------------------------
# Command-line input of user commands to display all regions
#
# -------------------------------------------------------------------

import sys, os, exceptions

import roslib; roslib.load_manifest('region_viewer')
import roslib.packages
import rospy
from region_msgs.msg import Regions as RRegions, Region as RRegion
from region_msgs.msg import VoronoiPoint2D as VPoint, Tag as RTag, Tags as RTags
from geometry_msgs.msg import Point as RPoint, Point32 as RPoint32, Polygon as RPolygon, Pose as RPose
import cv

class RegionDisplay:
    def __init__(self, regions_topic, tag_topic):
        print "Subscribing to regions topic: '{0}', and named points topic: '{1}'".format(regions_topic, tag_topic)
        self.m_regions_sub = rospy.Subscriber(regions_topic, RRegions,
                                              self._on_update_regions)
        self.m_namedpoints_sub = rospy.Subscriber(tag_topic, RTags,
                                                  self._on_update_namedpoints)
        self.m_mat = None
        self.m_white = (255,255,255)
        self.m_blue = (255,0,0)
        self.m_black = (0,0,0)
        self.m_green = (0,255,0)
        self.m_fontRegionScale = 0.55
        self.m_fontNPointScale = 0.35
        self.m_fontRegion = cv.InitFont(cv.CV_FONT_HERSHEY_COMPLEX,
                                        self.m_fontRegionScale, self.m_fontRegionScale)
        self.m_fontNPoint = cv.InitFont(cv.CV_FONT_HERSHEY_COMPLEX,
                                        self.m_fontNPointScale, self.m_fontNPointScale)
        self.m_regions = []
        self.m_namedpoints = []

    def _set_window_bounds(self, min_x, min_y, max_x, max_y):
        self.m_origin = RPoint()
        self.m_resolution = 0.04
        (self.m_height, self.m_origin.y) = _calculate_limits_y(min_y, max_y, 0.5,
                                                              self.m_resolution)
        (self.m_width, self.m_origin.x) = _calculate_limits_x(min_x, max_x, 0.5,
                                                               self.m_resolution)

        self.m_windowName = "Regions"
#        self.m_windowName = "Grouped Regions"
#        cv.NamedWindow(self.m_windowName, cv.CV_WINDOW_NORMAL | cv.CV_WINDOW_FREERATIO)
        cv.NamedWindow(self.m_windowName, cv.CV_WINDOW_NORMAL)
        self._resize_window()
#        cv.NamedWindow(self.m_windowName)

    def _rpoint2cvpoint(self, rpoint):
        newx = int(rpoint.x/self.m_resolution) - self.m_origin.x
        newy = self.m_height - 1 -(rpoint.y/self.m_resolution - self.m_origin.y)
        return (int(newx),int(newy))

    def _on_update_regions(self, data):
        self.m_regions = []
        self.m_regions = data.regions
        (min_x, min_y, max_x, max_y) = _calculate_bounds(data.regions)
        if not min_x:
            mix_x = 0.0; min_y = 0.0
            max_x = 100.0; max_y = 100.0
        self._set_window_bounds(min_x, min_y, max_x, max_y)
        self._update_image()

    def _on_update_namedpoints(self, data):
        self.m_namedpoints = []
        self.m_namedpoints = data.points
        self._update_image()

    def _resize_window(self):
        self.m_mat = cv.CreateMat(self.m_height, self.m_width, cv.CV_8UC3)
        self._update_image()

    # draw a white background
    def _draw_background(self):
        for y in range (0, self.m_height):
            for x in range (0, self.m_width):
                self.m_mat[y,x] = self.m_white

    def _draw_polygon(self, rpoints, color):
        num = len(rpoints)
        if num < 3: return
        first = self._rpoint2cvpoint(rpoints[0])

        next = first
        for i in range(1, num-1):
            cvp = self._rpoint2cvpoint(rpoints[i])
            cv.Line(self.m_mat, next, cvp, color)
            next = cvp
        if next != first:
            cv.Line(self.m_mat, next, first, color)

    def _draw_voronoi(self, voronoi, labels, vcolor, lcolor):
        if not voronoi: return
        maxd = voronoi[0].distance
        maxp = self._rpoint2cvpoint(voronoi[0].point)
        for v in voronoi:
            point = self._rpoint2cvpoint(v.point)
            cv.Circle(self.m_mat, point, 0, vcolor, 1)
            if v.distance > maxd:
                maxd = v.distance
                maxp = point

        label = ", ".join(labels)
        if label:
            textpoint = (maxp[0] - 10, maxp[1]-2)
            cv.PutText(self.m_mat, label, textpoint,
                       self.m_fontNPoint, lcolor)
        else:
            return
            cv.Circle(self.m_mat, maxp, 2, vcolor, 2)


    # Only call if the cv.matrix is valid
    def _draw_region(self, region):
#        midpoint = self._rpoint2cvpoint(region.midpoint)
#        if len(region.name) > 0:
#            textpoint = (midpoint[0]-25, midpoint[1]-5)
#            cv.PutText(self.m_mat, region.name, textpoint, self.m_fontRegion,
#                       self.m_black)
#            cv.Circle(self.m_mat, midpoint, 1, self.m_black, 2)
        if next((True for l in region.labels if l == 'office8'), False):
            self._draw_polygon(region.polygon.outer.points, self.m_black)
            self._draw_voronoi(region.voronoi, region.labels, self.m_black, self.m_green)
        else:
            self._draw_polygon(region.polygon.outer.points, self.m_blue)
            self._draw_voronoi(region.voronoi, region.labels, self.m_green, self.m_black)

    def _draw_namedpoint(self, namedpoint):
        if len(namedpoint.name) <= 0: return
        point = self._rpoint2cvpoint(namedpoint.point)
        textpoint = (point[0] - 10, point[1]-5)
        cv.PutText(self.m_mat, namedpoint.name, textpoint, self.m_fontNPoint,
                   self.m_green)
        cv.Circle(self.m_mat, point, 1, self.m_green, 2)

    def _update_image(self):
        if self.m_mat is None: return
        self._draw_background()
        for namedpoint in self.m_namedpoints:
            self._draw_namedpoint(namedpoint)
        for region in self.m_regions:
            self._draw_region(region)
        cv.ShowImage(self.m_windowName, self.m_mat)


#---------------------------------------------------------
# Calculate the bounds of a set of regions
#---------------------------------------------------------
def _calculate_bounds(regions):
    min_x = None; max_x = None; min_y = None; max_y = None
    for region in regions:
        for point in region.polygon.outer.points:
            if min_x is None:
                min_x = point.x; max_x = point.x
                min_y = point.y; max_y = point.y
            else:
                if point.x < min_x: min_x = point.x
                if point.x > max_x: max_x = point.x
                if point.y < min_y: min_y = point.y
                if point.y > max_y: max_y = point.y
    return (min_x, min_y, max_x, max_y)

#---------------------------------------------------------
# Calculate the width and origin for an x dimension
#---------------------------------------------------------
def _calculate_limits_x(min_d, max_d, buf, resolution):
    width = (max_d - min_d) + buf*2.0
    if min_d >= 0 and max_d >= 0:
        origin = min_d+buf
    else:
        origin = min_d-buf
    return (int(width/resolution), int(origin/resolution))

#---------------------------------------------------------
# Calculate the width and origin for a y dimension
#---------------------------------------------------------
def _calculate_limits_y(min_d, max_d, buf, resolution):
    width = (max_d - min_d) + buf*2.0
    if min_d >= 0 and max_d >= 0:
        origin = min_d+buf
    else:
        origin = min_d-buf
    return (int(width/resolution), int(origin/resolution))


#---------------------------------------------------------
# main
#---------------------------------------------------------

if __name__ == '__main__':

    try:
        rospy.init_node('show_regions', log_level=rospy.DEBUG)

        # Displays the high-level grouped regions
#        regions_topic = rospy.get_param('regions', 'regions_grouped')

        # Displays the low-level regions
        regions_topic = rospy.get_param('regions', 'regions')

        tags_topic = rospy.get_param('tags', 'tags')
        rd = RegionDisplay(regions_topic, tags_topic)
        done = False
        while not rospy.is_shutdown():
            cv.WaitKey(2)

    except:
        print "Unexpected error", sys.exc_info()[0]
        raise
    cv.DestroyAllWindows()
