#!/usr/bin/env python

#--------------------------------------------------------------------
# Command-line input of user commands to display all regions
#
# -------------------------------------------------------------------

import sys, os, exceptions

import roslib; roslib.load_manifest('region_viewer')
import roslib.packages
import rospy
from region_msgs.msg import Regions, Region, Tag, Tags, Point2D
from region_msgs.msg import SimplePolygon2D, PolygonWithHoles2D
from geometry_msgs.msg import Pose as RPose
from nav_msgs.msg import OccupancyGrid, MapMetaData
import cv

class RegionDisplay:
    def __init__(self, map_topic, regions_serv, namedpoints_topic):
        print "Subscribing to map topic: '{0}', and regions service: '{1}', and named points topic: '{2}'".format(map_topic, regions_serv, namedpoints_topic)
        self.m_width = 0
        self.m_height = 0
        self.m_resolution = 0.0
        self.m_origin = Point2D()
        self.m_regions = []
        self.m_namedpoints = []
        self.m_mat = None
        self.m_white = (255,255,255)
        self.m_blue = (255,0,0)
        self.m_black = (0,0,0)
        self.m_green = (0,255,0)
        self.m_fontRegionScale = 0.45
        self.m_fontNPointScale = 0.30
        self.m_fontRegion = cv.InitFont(cv.CV_FONT_HERSHEY_COMPLEX,
                                        self.m_fontRegionScale, self.m_fontRegionScale)
        self.m_fontNPoint = cv.InitFont(cv.CV_FONT_HERSHEY_COMPLEX, self.m_fontNPointScale, self.m_fontNPointScale)

        self.m_windowName = "Regions"
#        self.m_windowName = "Grouped Regions"
#        cv.NamedWindow(self.m_windowName, cv.CV_WINDOW_NORMAL | cv.CV_WINDOW_FREERATIO)
        cv.NamedWindow(self.m_windowName)
        self.m_map_sub = rospy.Subscriber(map_topic, OccupancyGrid, self._on_update_map)
        self.m_regions_sub = rospy.Subscriber(regions_serv, Regions, self._on_update_regions)
        self.m_namedpoints_sub = rospy.Subscriber(namedpoints_topic, Tags, self._on_update_namedpoints)

    def _rpoint2cvpoint(self, rpoint):
        newx = (rpoint.x - self.m_origin.x)/self.m_resolution
        newy = self.m_height - 1 -(rpoint.y - self.m_origin.y)/self.m_resolution
        return (int(newx),int(newy))

    def _on_update_regions(self, data):
        self.m_regions = []
        self.m_regions = data.regions
        self._update_image()

    def _on_update_namedpoints(self, data):
        self.m_namedpoints = []
        self.m_namedpoints = data.points
        self._update_image()

    def _on_update_map(self, data):
        changed = False
        if self.m_height != data.info.height:
            self.m_height = data.info.height
            changed = True
        if self.m_width != data.info.width:
            self.m_width = data.info.width
            changed = True
        if self.m_resolution != data.info.resolution:
            self.m_resolution = data.info.resolution
            changed = True
        if self.m_origin != data.info.origin.position:
            self.m_origin = data.info.origin.position
            changed = True
        if changed:
            self.m_mat = cv.CreateMat(self.m_height, self.m_width, cv.CV_8UC3)
        self._update_image()
#                self.m_mat[self.m_height-1 - x]

    # draw a white background
    def _draw_background(self):
        for x in range (0, self.m_height-1):
            for y in range (0, self.m_width-1):
                self.m_mat[x,y] = self.m_white

    def _draw_polygon(self, rpoints):
        num = len(rpoints)
        if num < 3: return
        first = self._rpoint2cvpoint(rpoints[0])
        next = first
        for i in range(1, num-1):
            cvp = self._rpoint2cvpoint(rpoints[i])
            cv.Line(self.m_mat, next, cvp, self.m_blue)
            next = cvp
        if next != first:
            cv.Line(self.m_mat, next, first, self.m_blue)

    # Only call if the cv.matrix is valid
    def _draw_region(self, region):
#        midpoint = self._rpoint2cvpoint(region.midpoint)
#        if len(region.name) > 0:
#            textpoint = (midpoint[0]-25, midpoint[1]-5)
#            cv.PutText(self.m_mat, region.name, textpoint, self.m_fontRegion, self.m_black)
##            cv.Circle(self.m_mat, midpoint, 1, self.m_black, 2)
        self._draw_polygon(region.polygon.outer.points)

    def _draw_namedpoint(self, namedpoint):
        if len(namedpoint.name) <= 0: return
        point = self._rpoint2cvpoint(namedpoint.point)
        textpoint = (point[0] - 10, point[1]-5)
        cv.PutText(self.m_mat, namedpoint.name, textpoint, self.m_fontNPoint, self.m_green)
        cv.Circle(self.m_mat, point, 1, self.m_green, 2)

    def _update_image(self):
        if self.m_mat is None: return
        self._draw_background()
        for namedpoint in self.m_namedpoints:
            self._draw_namedpoint(namedpoint)
        for region in self.m_regions:
            self._draw_region(region)
        cv.ShowImage(self.m_windowName, self.m_mat)

if __name__ == '__main__':

    try:
        rospy.init_node('show_regions', log_level=rospy.DEBUG)
        map_topic = rospy.get_param('map','map')

        # Displays the high-level grouped regions
#        regions_serv = rospy.get_param('regions_grouped', 'regions_grouped')

        # Displays the low-level regions
        regions_serv = rospy.get_param('regions_grouped', 'regions')

        namedpoints_topic = rospy.get_param('named_points', 'named_points')
        rd = RegionDisplay(map_topic, regions_serv, namedpoints_topic)
        done = False
        while not rospy.is_shutdown():
            cv.WaitKey(2)

    except:
        print "Unexpected error", sys.exc_info()[0]
        raise
    cv.DestroyAllWindows()
